# differential_equations.py

################
# COFe: O2 #####
################
import gmpy2 as gp


def COFe_O2(xg_O2_, values):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_O2 for the C-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        values (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        KD1_,
        KD2,
        y,
        y_CO2_,
        y_O2_,
        y_CO_,
        M_C,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = values
    result = (
        (
            -K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            + wt_C / M_C
            - K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
        )
        * (
            K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            * (
                K2_
                * M_CO2
                * P
                * y_CO2_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                - M_m_
                * (
                    K2_
                    * P
                    * y_CO2_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    - 0.5
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_CO2_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_CO_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_CO2_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
            )
            - K2_
            * P
            * y_CO2_
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            - (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            * (
                M_CO2
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                - M_O2
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
                - 0.5
                * M_CO2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_CO2_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - 0.5
                * M_CO
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_CO2_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_CO_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            / (
                M_CO2
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                -1
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_CO2_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
                - 0.5
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_CO2_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_CO_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            / (
                M_CO2
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            + 0.5
            * K2_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_CO2_**2
            * y_O2_ ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * y_CO_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                ** 2
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            - K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * xg_O2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_CO2_**2
            * y_O2_ ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_**2
                * M_CO
                * y_CO_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                ** 2
            )
        )
        / (
            -K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            + (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            / (
                M_CO2
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            - K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
        )
        ** 2
        - 0.25
        * (
            -2.0
            * K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
        )
        * (
            2.0
            * K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            * (
                K2_
                * M_CO2
                * P
                * y_CO2_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                - M_m_
                * (
                    K2_
                    * P
                    * y_CO2_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    - 0.5
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_CO2_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_CO_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_CO2_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
            )
            - 2.0
            * K2_
            * P
            * y_CO2_
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            - (
                2.0 * xg_O2_
                + 2.0
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            * (
                M_CO2
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                - M_O2
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
                - 0.5
                * M_CO2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_CO2_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - 0.5
                * M_CO
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_CO2_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_CO_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            / (
                M_CO2
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                2.0
                - 2.0
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + 1.0
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_CO2_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
                - 0.5
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_CO2_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_CO_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            / (
                M_CO2
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            + 1.0
            * K2_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_CO2_**2
            * y_O2_ ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * y_CO_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                ** 2
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            - K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * xg_O2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_CO2_**2
            * y_O2_ ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_**2
                * M_CO
                * y_CO_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                ** 2
            )
        )
        / (
            -K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            + 0.5
            * (
                2.0 * xg_O2_
                + 2.0
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            / (
                M_CO2
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            - 0.5
            * K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
        )
        ** 2
        + (
            -K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            * (
                K2_
                * M_CO2
                * P
                * y_CO2_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                - M_m_
                * (
                    K2_
                    * P
                    * y_CO2_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    - 0.5
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_CO2_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_CO_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_CO2_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
            )
            + K2_
            * P
            * y_CO2_
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            - 0.5
            * K2_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_CO2_**2
            * y_O2_ ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * y_CO_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                ** 2
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            + K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * xg_O2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_CO2_**2
            * y_O2_ ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_**2
                * M_CO
                * y_CO_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                ** 2
            )
        )
        / (
            -K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            + (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            / (
                M_CO2
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            - K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
        )
        - (
            -2.0
            * K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            * (
                K2_
                * M_CO2
                * P
                * y_CO2_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                - M_m_
                * (
                    K2_
                    * P
                    * y_CO2_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    - 0.5
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_CO2_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_CO_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_CO2_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
            )
            + 2.0
            * K2_
            * P
            * y_CO2_
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
            - 1.0
            * K2_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_CO2_**2
            * y_O2_ ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * y_CO_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                ** 2
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            + K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * xg_O2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_CO2_**2
            * y_O2_ ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_**2
                * M_CO
                * y_CO_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                ** 2
            )
        )
        / (
            -2.0
            * K2_
            * P
            * y_CO2_
            * (1.0 - xg_O2_)
            / (
                (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                * (
                    K2_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K2_
                        * P
                        * y_CO2_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_CO2_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
            )
            + (
                2.0 * xg_O2_
                + 2.0
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            / (
                M_CO2
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_CO_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_CO2_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_CO_)
                    )
                )
            )
            - K3_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_CO
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_CO2_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_CO_)
                )
            )
        )
    )
    return result


################
# HOFe: O2 #####
################

# no insolubles


def HOFe_O2(xg_O2_, constants):
    """Evaluate differential of wtg_H - wtg_O with respect to xg_O2 for the H-O-Fe
    system with no dissolved H2 in the melt.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,  # noqa
        KD1_,
        KD2,
        y,
        y_H2O_,
        y_O2_,
        y_H2_,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            -(
                (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (-xg_O2_ + 1.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K2_ * P * y_H2O_ * (-xg_O2_ + 1.0))
                + M_m_
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (-xg_O2_ + 1.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K2_ * P * y_H2O_ * (-xg_O2_ + 1.0))
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            * (
                M_H2O
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                - M_O2
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * M_H2O
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * M_H2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                -1
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * P
                * y_H2O_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.25
                * K2_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                * (
                    M_H2O
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (-xg_O2_ + 1.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K2_
                                * P
                                * y_H2O_
                                * (-xg_O2_ + 1.0)
                                / (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
            )
        )
        / (
            -(
                (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            / (
                M_H2O
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        + (
            -(
                (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            * (
                -M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (-xg_O2_ + 1.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K2_ * P * y_H2O_ * (-xg_O2_ + 1.0))
                + M_m_
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (-xg_O2_ + 1.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K2_ * P * y_H2O_ * (-xg_O2_ + 1.0))
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * P
                * y_H2O_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.25
                * K2_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                * (
                    M_H2O
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (-xg_O2_ + 1.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K2_
                                * P
                                * y_H2O_
                                * (-xg_O2_ + 1.0)
                                / (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
            )
        )
        / (
            -(
                (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            / (
                M_H2O
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
        )
        - (
            -(
                (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            * (
                -M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (-xg_O2_ + 1.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K2_ * P * y_H2O_ * (-xg_O2_ + 1.0))
                + M_m_
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (-xg_O2_ + 1.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K2_ * P * y_H2O_ * (-xg_O2_ + 1.0))
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (-2.0 * y + 1.0)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (-2.0 * y + 1.0)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
            - (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * P
                * y_H2O_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.25
                * K2_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                * (
                    M_H2O
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (-xg_O2_ + 1.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K2_
                                * P
                                * y_H2O_
                                * (-xg_O2_ + 1.0)
                                / (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
            )
        )
        / (
            -(
                (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                2.0 * xg_O2_
                + (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            / (
                M_H2O
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
        )
        - (
            -(
                (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (-xg_O2_ + 1.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K2_ * P * y_H2O_ * (-xg_O2_ + 1.0))
                + M_m_
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (-xg_O2_ + 1.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K2_ * P * y_H2O_ * (-xg_O2_ + 1.0))
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * xg_O2_
                + (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            * (
                M_H2O
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                - M_O2
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * M_H2O
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * M_H2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                2.0
                - 1
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * P
                * y_H2O_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.25
                * K2_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                K2_
                * P
                * y_H2O_
                * (-xg_O2_ + 1.0)
                * (
                    M_H2O
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (-xg_O2_ + 1.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K2_
                                * P
                                * y_H2O_
                                * (-xg_O2_ + 1.0)
                                / (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
            )
        )
        / (
            -(
                (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K2_
                    * P
                    * y_H2O_
                    * (-xg_O2_ + 1.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K2_
                            * P
                            * y_H2O_
                            * (-xg_O2_ + 1.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                2.0 * xg_O2_
                + (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            / (
                M_H2O
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
    )
    return result


# insolubles included - H2
def HOFe_O2_2(xg_O2_, constants):
    """Evaluate differential of wtg_H - wtg_O with respect to xg_O2 for the H-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_H,
        wt_Fe,
        K1_,
        K2_,  # noqa
        K3_,
        KD1_,
        KD2,
        y,
        y_H2O_,
        y_O2_,
        y_H2_,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            -(
                (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + 0.5 * wt_H / M_H
            - K3_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_H2
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
        )
        * (
            (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K3_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K3_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K3_ * P * y_H2O_ * (1.0 - xg_O2_))
                + M_m_
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K3_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K3_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K3_ * P * y_H2O_ * (1.0 - xg_O2_))
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            * (
                M_H2O
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                - M_O2
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * M_H2O
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * M_H2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                -1
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K3_
                * P
                * y_H2O_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.25
                * K3_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (
                    M_H2O
                    * (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K3_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
            )
            - K3_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            / (
                K1_
                * M_H2
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K3_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_H2
                * xg_O2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K3_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_H2O_**2
            * y_O2_ ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_**2
                * M_H2
                * y_H2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                ** 2
            )
        )
        / (
            -(
                (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            - K3_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_H2
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
        )
        ** 2
        + (
            -(
                (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            * (
                -M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K3_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K3_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K3_ * P * y_H2O_ * (1.0 - xg_O2_))
                + M_m_
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K3_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K3_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K3_ * P * y_H2O_ * (1.0 - xg_O2_))
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K3_
                * P
                * y_H2O_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.25
                * K3_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (
                    M_H2O
                    * (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K3_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
            )
            + K3_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            / (
                K1_
                * M_H2
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K3_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_H2
                * xg_O2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K3_
            * P**0.5
            * xg_O2_ ** (-1.5)
            * y_H2O_**2
            * y_O2_ ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_**2
                * M_H2
                * y_H2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                ** 2
            )
        )
        / (
            -(
                (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            - K3_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (1.0 - xg_O2_)
            / (
                K1_
                * M_H2
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
        )
        - (
            -(
                (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            * (
                -M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K3_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K3_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K3_ * P * y_H2O_ * (1.0 - xg_O2_))
                + M_m_
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K3_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K3_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K3_ * P * y_H2O_ * (1.0 - xg_O2_))
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
            - (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K3_
                * P
                * y_H2O_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.25
                * K3_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (
                    M_H2O
                    * (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K3_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
            )
        )
        / (
            -(
                (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                2.0 * xg_O2_
                + (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
        )
        - (
            -(
                (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K3_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K3_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K3_ * P * y_H2O_ * (1.0 - xg_O2_))
                + M_m_
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                * (
                    -0.5
                    * K3_
                    * P
                    * y_H2O_
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * K3_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                / (K3_ * P * y_H2O_ * (1.0 - xg_O2_))
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                2.0 * xg_O2_
                + (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            * (
                M_H2O
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                - M_O2
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * M_H2O
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * M_H2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                2.0
                - 1
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K3_
                * P
                * y_H2O_
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.25
                * K3_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                K3_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (
                    M_H2O
                    * (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K3_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
            )
        )
        / (
            -(
                (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K3_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K3_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                2.0 * xg_O2_
                + (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
    )
    return result


def HOFe_O2_xenia(xg_O2_, constants):
    """Evaluate differential of wtg_H - wtg_O with respect to xg_O2 for Xenia's H-O-Fe system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4a_,
        K4b_,
        K5_,
        y_H2O_,
        y_O2_,
        y_H2_,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -K4a_
            * (
                P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** K4b_
            / M_H2
            - (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            K1_
            * K4a_
            * K4b_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** K4b_
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (M_H2 * P * y_H2O_ * (1.0 - xg_O2_))
            - (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            * (
                M_H2O
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                - M_O2
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * M_H2O
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * M_H2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                -1
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            + (
                0.25
                * K2_
                * P
                * y_H2O_
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K2_
                    * P
                    * y_H2O_
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.125
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    * (
                        -2.0
                        * K2_
                        * P
                        * y_H2O_
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 1.0
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                    - 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                - 0.125
                * K2_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                -M_H2O
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            -K2_
                            * P
                            * y_H2O_
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 0.5
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.125
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        * (
                            -2.0
                            * K2_
                            * P
                            * y_H2O_
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 1.0
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                        - 0.25
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    - 0.125
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                - M_m_
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            -K2_
                            * P
                            * y_H2O_
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 0.5
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.125
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        * (
                            -2.0
                            * K2_
                            * P
                            * y_H2O_
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 1.0
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                        - 0.25
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    + 0.125
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
            )
            * (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4a_
            * (
                P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** K4b_
            / M_H2
            + (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            - (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        ** 2
        + (
            -K1_
            * K4a_
            * K4b_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** K4b_
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                / (K1_ * y_H2_)
            )
            * (
                -P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (M_H2 * P * y_H2O_ * (1.0 - xg_O2_))
            - (
                0.25
                * K2_
                * P
                * y_H2O_
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K2_
                    * P
                    * y_H2O_
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.125
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    * (
                        -2.0
                        * K2_
                        * P
                        * y_H2O_
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 1.0
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                    - 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                - 0.125
                * K2_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                -M_H2O
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            -K2_
                            * P
                            * y_H2O_
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 0.5
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.125
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        * (
                            -2.0
                            * K2_
                            * P
                            * y_H2O_
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 1.0
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                        - 0.25
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    - 0.125
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                - M_m_
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            -K2_
                            * P
                            * y_H2O_
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 0.5
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.125
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        * (
                            -2.0
                            * K2_
                            * P
                            * y_H2O_
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 1.0
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                        - 0.25
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    + 0.125
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
            )
            * (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4a_
            * (
                P
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** K4b_
            / M_H2
            + (
                (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            - (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        - (
            -(
                0.25
                * K2_
                * P
                * y_H2O_
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K2_
                    * P
                    * y_H2O_
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.125
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    * (
                        -2.0
                        * K2_
                        * P
                        * y_H2O_
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 1.0
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                    - 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                - 0.125
                * K2_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                -M_H2O
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            -K2_
                            * P
                            * y_H2O_
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 0.5
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.125
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        * (
                            -2.0
                            * K2_
                            * P
                            * y_H2O_
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 1.0
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                        - 0.25
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    - 0.125
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                - M_m_
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            -K2_
                            * P
                            * y_H2O_
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 0.5
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.125
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        * (
                            -2.0
                            * K2_
                            * P
                            * y_H2O_
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 1.0
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                        - 0.25
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    + 0.125
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
            )
            * (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.147
            * wt_Fe
            * (P * xg_O2_ * y_O2_) ** 0.196
            * gp.exp(K5_)
            / (
                M_Fe
                * xg_O2_
                * (0.5 * (P * xg_O2_ * y_O2_) ** 0.196 * gp.exp(K5_) + 1.0)
            )
            + 0.098
            * wt_Fe
            * (P * xg_O2_ * y_O2_) ** 0.196
            * (0.75 * (P * xg_O2_ * y_O2_) ** 0.196 * gp.exp(K5_) + 1.0)
            * gp.exp(K5_)
            / (
                M_Fe
                * xg_O2_
                * (0.5 * (P * xg_O2_ * y_O2_) ** 0.196 * gp.exp(K5_) + 1.0) ** 2
            )
        )
        / (
            (
                2.0 * xg_O2_
                + (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            - (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        - (
            -(
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (0.75 * (P * xg_O2_ * y_O2_) ** 0.196 * gp.exp(K5_) + 1.0)
            / (M_Fe * (0.5 * (P * xg_O2_ * y_O2_) ** 0.196 * gp.exp(K5_) + 1.0))
        )
        * (
            -(
                2.0 * xg_O2_
                + (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            * (
                M_H2O
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                - M_O2
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * M_H2O
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * M_H2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_**2
                    * y_H2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                2.0
                - 1
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_H2O_
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            + (
                0.25
                * K2_
                * P
                * y_H2O_
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K2_
                    * P
                    * y_H2O_
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.125
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    * (
                        -2.0
                        * K2_
                        * P
                        * y_H2O_
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 1.0
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                    - 0.25
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                - 0.125
                * K2_
                * P**0.5
                * xg_O2_ ** (-1.5)
                * y_H2O_**2
                * y_O2_ ** (-0.5)
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    ** 2
                )
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                -M_H2O
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            -K2_
                            * P
                            * y_H2O_
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 0.5
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.125
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        * (
                            -2.0
                            * K2_
                            * P
                            * y_H2O_
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 1.0
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                        - 0.25
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    - 0.125
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
                - M_m_
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            -K2_
                            * P
                            * y_H2O_
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 0.5
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.5
                        * K2_
                        * P
                        * y_H2O_
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.125
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        * (
                            -2.0
                            * K2_
                            * P
                            * y_H2O_
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            + 1.0
                            * K2_
                            * P**0.5
                            * xg_O2_ ** (-1.5)
                            * y_H2O_**2
                            * y_O2_ ** (-0.5)
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                K1_
                                * y_H2_
                                * (
                                    1.0
                                    + P ** (-0.5)
                                    * xg_O2_ ** (-0.5)
                                    * y_H2O_
                                    * y_O2_ ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2
                            )
                        )
                        / (K2_ * P * y_H2O_ * (1.0 - xg_O2_) * (K3_ - 4.0))
                        - 0.25
                        * K2_
                        * P**0.5
                        * xg_O2_ ** (-1.5)
                        * y_H2O_**2
                        * y_O2_ ** (-0.5)
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            K1_
                            * y_H2_
                            * (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2
                        )
                    )
                    + 0.125
                    * K2_
                    * P**0.5
                    * xg_O2_ ** (-1.5)
                    * y_H2O_**2
                    * y_O2_ ** (-0.5)
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        K1_
                        * y_H2_
                        * (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                )
            )
            * (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            (
                2.0 * xg_O2_
                + (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
            )
            / (
                M_H2O
                * (1.0 - xg_O2_)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (1.0 - xg_O2_)
                / (
                    K1_
                    * y_H2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                )
            )
            - (
                -0.25
                * K2_
                * P
                * y_H2O_
                * (1.0 - xg_O2_)
                * (K3_ - 4.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_H2O_
                    * y_O2_ ** (-0.5)
                    / (K1_ * y_H2_)
                )
                + 0.5
                * (
                    -K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (4.0 - K3_)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        - K3_
                    )
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.25
                    * (
                        K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (K3_ - 4.0)
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                    )
                    ** 2.0
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    -0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    + 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                )
                + M_m_
                * (
                    0.25
                    * K2_
                    * P
                    * y_H2O_
                    * (1.0 - xg_O2_)
                    * (K3_ - 4.0)
                    / (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_H2O_
                        * y_O2_ ** (-0.5)
                        / (K1_ * y_H2_)
                    )
                    - 0.5
                    * (
                        -K2_
                        * P
                        * y_H2O_
                        * (1.0 - xg_O2_)
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (4.0 - K3_)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            - K3_
                        )
                        / (
                            1.0
                            + P ** (-0.5)
                            * xg_O2_ ** (-0.5)
                            * y_H2O_
                            * y_O2_ ** (-0.5)
                            / (K1_ * y_H2_)
                        )
                        + 0.25
                        * (
                            K2_
                            * P
                            * y_H2O_
                            * (1.0 - xg_O2_)
                            * (K3_ - 4.0)
                            / (
                                1.0
                                + P ** (-0.5)
                                * xg_O2_ ** (-0.5)
                                * y_H2O_
                                * y_O2_ ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 2.0
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        ** 2
    )
    return result


################
# SOFe: O2 #####
################


def SOFe_O2(xg_O2_, constants):
    """Evaluate differential of wtg_S - wtg_O with respect to xg_O2 for the S-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        KD1_,
        KD2,
        y,
        y_SO2_,
        y_O2_,
        y_S2_,
        M_S,
        M_O,
        M_Fe,
        M_O2,
        M_S2,
        M_SO2,
        M_SO3,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,  # noqa
    ) = constants
    result = (
        (
            wt_S / M_S
            - P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S)
            - K2_
            * y_SO2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-xg_O2_ + 1.0)
            / (
                K1_
                * M_S
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
        )
        * (
            K1_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_O2_ ** (-0.5)
                * y_SO2_
                / (K1_ * y_S2_)
            )
            * (
                -0.5
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                - 0.25
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                + 0.25
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_**2
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
            )
            / (K3_ * M_S * y_SO2_ * (-xg_O2_ + 1.0))
            - (
                (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + 2.0
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            * (
                -M_O2
                + M_SO2
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                + 0.5
                * M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                - 0.5
                * M_SO2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
                - 0.5
                * M_S2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_**2
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_S2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
            )
            / (
                M_O2 * xg_O2_
                + M_SO2
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 2
            - (
                -1
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + 0.5
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
                - 2.0
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                - 1.0
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                + 1.0
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_**2
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_S2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
            )
            / (
                M_O2 * xg_O2_
                + M_SO2
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            + 1.5
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S * xg_O2_)
            - K2_
            * y_SO2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            / (
                K1_
                * M_S
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
            - 1.0
            * K2_
            * y_SO2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-xg_O2_ + 1.0)
            / (
                K1_
                * M_S
                * xg_O2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
            + 0.5
            * K2_
            * P ** (-0.5)
            * xg_O2_ ** (-1.5)
            * y_O2_ ** (-0.5)
            * y_SO2_**2
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-xg_O2_ + 1.0)
            / (
                K1_**2
                * M_S
                * y_S2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                ** 2
            )
        )
        / (
            (
                (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + 2.0
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            / (
                M_O2 * xg_O2_
                + M_SO2
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            - P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S)
            - K2_
            * y_SO2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-xg_O2_ + 1.0)
            / (
                K1_
                * M_S
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
        )
        ** 2
        + (
            -K1_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_O2_ ** (-0.5)
                * y_SO2_
                / (K1_ * y_S2_)
            )
            * (
                -0.5
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                - 0.25
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                + 0.25
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_**2
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
            )
            / (K3_ * M_S * y_SO2_ * (-xg_O2_ + 1.0))
            - 1.5
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S * xg_O2_)
            + K2_
            * y_SO2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            / (
                K1_
                * M_S
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
            + 1.0
            * K2_
            * y_SO2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-xg_O2_ + 1.0)
            / (
                K1_
                * M_S
                * xg_O2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
            - 0.5
            * K2_
            * P ** (-0.5)
            * xg_O2_ ** (-1.5)
            * y_O2_ ** (-0.5)
            * y_SO2_**2
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-xg_O2_ + 1.0)
            / (
                K1_**2
                * M_S
                * y_S2_
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                ** 2
            )
        )
        / (
            (
                (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + 2.0
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            / (
                M_O2 * xg_O2_
                + M_SO2
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            - P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S)
            - K2_
            * y_SO2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-xg_O2_ + 1.0)
            / (
                K1_
                * M_S
                * (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
        )
        - (
            -3.0
            * K1_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_O2_ ** (-0.5)
                * y_SO2_
                / (K1_ * y_S2_)
            )
            * (
                -0.5
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                - 0.25
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                + 0.25
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_**2
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
            )
            / (K3_ * M_S * y_SO2_ * (-xg_O2_ + 1.0))
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (-2.0 * y + 1.0)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (-2.0 * y + 1.0)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
            - 4.5
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S * xg_O2_)
        )
        / (
            (
                2.0 * xg_O2_
                + 2.0
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
            / (
                M_O2 * xg_O2_
                + M_SO2
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            - 3.0
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S)
        )
        - (
            wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - 3.0
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S)
        )
        * (
            3.0
            * K1_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            * (
                1.0
                + P ** (-0.5)
                * xg_O2_ ** (-0.5)
                * y_O2_ ** (-0.5)
                * y_SO2_
                / (K1_ * y_S2_)
            )
            * (
                -0.5
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                - 0.25
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                + 0.25
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_**2
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
            )
            / (K3_ * M_S * y_SO2_ * (-xg_O2_ + 1.0))
            - (
                2.0 * xg_O2_
                + 2.0
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
            * (
                -M_O2
                + M_SO2
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                + 0.5
                * M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * xg_O2_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
                - 0.5
                * M_SO2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
                - 0.5
                * M_S2
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_**2
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_**2
                    * y_S2_**2
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
            )
            / (
                M_O2 * xg_O2_
                + M_SO2
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 2
            - (
                2.0
                - 2.0
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + 1.0
                * P ** (-0.5)
                * xg_O2_ ** (-1.5)
                * y_O2_ ** (-0.5)
                * y_SO2_
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                    ** 2
                )
            )
            / (
                M_O2 * xg_O2_
                + M_SO2
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            + 4.5
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S * xg_O2_)
        )
        / (
            (
                2.0 * xg_O2_
                + 2.0
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
            )
            / (
                M_O2 * xg_O2_
                + M_SO2
                * (-xg_O2_ + 1.0)
                / (
                    1.0
                    + P ** (-0.5)
                    * xg_O2_ ** (-0.5)
                    * y_O2_ ** (-0.5)
                    * y_SO2_
                    / (K1_ * y_S2_)
                )
                + M_S2
                * y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * y_S2_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            - 3.0
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_SO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-xg_O2_ + 1.0)
                / (
                    K1_
                    * (
                        1.0
                        + P ** (-0.5)
                        * xg_O2_ ** (-0.5)
                        * y_O2_ ** (-0.5)
                        * y_SO2_
                        / (K1_ * y_S2_)
                    )
                )
            )
            ** 0.5
            / (K3_ * M_S)
        )
        ** 2
    )
    return result


#####################
# SHOFe: O2, S2 #####
#####################


def SHOFe_SO_O2(xg_O2_, xg_S2_, constants):
    """Evaluate differential of wtg_S - wtg_O with respect to xg_O2 for the S-H-O-Fe
    system with no dissolved H2 or H2S.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_S,
        wt_H,  # noqa
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        KD1_,
        KD2,
        y,
        y_S2_,
        y_O2_,
        y_SO2_,
        y_H2S_,
        y_H2O_,
        y_H2_,
        M_H,  # noqa
        M_S,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_S2,
        M_SO2,
        M_SO3,  # noqa
        M_H2S,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        + (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            -0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            * (
                -K2_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_H2O
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                - M_O2
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -4.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (-2.0 * y + 1.0)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (-2.0 * y + 1.0)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K4_
                                * P
                                * y_H2O_
                                * (
                                    -K2_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    - xg_O2_
                                    - xg_S2_
                                    + 1.0
                                )
                                / (
                                    K3_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        - (
            -3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            4.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                -K2_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_H2O
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                - M_O2
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                2.0 * K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                + 2.0
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K4_
                                * P
                                * y_H2O_
                                * (
                                    -K2_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    - xg_O2_
                                    - xg_S2_
                                    + 1.0
                                )
                                / (
                                    K3_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
    )
    return result


def SHOFe_SO_S2(xg_O2_, xg_S2_, constants):
    """Evaluate differential of wtg_S - wtg_O with respect to xg_S2 for the S-H-O-Fe system with no dissolved H2 or H2S.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_S,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        KD1_,
        KD2,
        y,
        y_S2_,
        y_O2_,
        y_SO2_,
        y_H2S_,
        y_H2O_,
        y_H2_,
        M_H,
        M_S,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        + (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            * (
                -0.5
                * K2_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K3_**2
                * M_H2S
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K3_
                * M_H2O
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_S2
                + 0.5
                * K3_
                * M_H2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 0.5
                * K3_**2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K3_
                    * K4_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K3_
                    * K4_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K3_
                * K4_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K4_
                                * P
                                * y_H2O_
                                * (
                                    -K2_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    - xg_O2_
                                    - xg_S2_
                                    + 1.0
                                )
                                / (
                                    K3_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        - (
            -3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K3_
                    * K4_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K3_
                    * K4_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                -0.5
                * K2_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K3_**2
                * M_H2S
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K3_
                * M_H2O
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_S2
                + 0.5
                * K3_
                * M_H2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 0.5
                * K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K3_
                * K4_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K4_
                                * P
                                * y_H2O_
                                * (
                                    -K2_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    - xg_O2_
                                    - xg_S2_
                                    + 1.0
                                )
                                / (
                                    K3_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
    )
    return result


def SHOFe_SH_O2(xg_O2_, xg_S2_, constants):
    """Evaluate differential of wtg_S - wtg_H with respect to xg_O2 for the S-H-O-Fe system with no dissolved H2 or H2S.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_S,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        KD1_,
        KD2,
        y,
        y_S2_,
        y_O2_,
        y_SO2_,
        y_H2S_,
        y_H2O_,
        y_H2_,
        M_H,
        M_S,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -(
                (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            * (
                -K2_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_H2O
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                - M_O2
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K4_
                                * P
                                * y_H2O_
                                * (
                                    -K2_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    - xg_O2_
                                    - xg_S2_
                                    + 1.0
                                )
                                / (
                                    K3_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -(
                (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        + (
            0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        + (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            -0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            * (
                -K2_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_H2O
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                - M_O2
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -(
                (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K4_
                                * P
                                * y_H2O_
                                * (
                                    -K2_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    - xg_O2_
                                    - xg_S2_
                                    + 1.0
                                )
                                / (
                                    K3_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -(
                (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
    )
    return result


def SHOFe_SH_S2(xg_O2_, xg_S2_, constants):
    """Evaluate differential of wtg_S - wtg_H with respect to xg_S2 for the S-H-O-Fe system with no dissolved H2 or H2S.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_S,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        KD1_,
        KD2,
        y,
        y_S2_,
        y_O2_,
        y_SO2_,
        y_H2S_,
        y_H2O_,
        y_H2_,
        M_H,
        M_S,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -(
                (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K3_
                    * K4_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K3_
                    * K4_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            * (
                -0.5
                * K2_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K3_**2
                * M_H2S
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K3_
                * M_H2O
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_S2
                + 0.5
                * K3_
                * M_H2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                -0.5
                * K3_**2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - 0.5
                * K3_
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K3_
                * K4_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K4_
                                * P
                                * y_H2O_
                                * (
                                    -K2_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    - xg_O2_
                                    - xg_S2_
                                    + 1.0
                                )
                                / (
                                    K3_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -(
                (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        + (
            -0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        + (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            * (
                -0.5
                * K2_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K3_**2
                * M_H2S
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K3_
                * M_H2O
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_S2
                + 0.5
                * K3_
                * M_H2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 0.5
                * K3_**2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -(
                (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K3_
                    * K4_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K3_
                    * K4_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            ** 2
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K3_
                * K4_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K3_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -(
                            (
                                K4_
                                * P
                                * y_H2O_
                                * (
                                    -K2_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    - xg_O2_
                                    - xg_S2_
                                    + 1.0
                                )
                                / (
                                    K3_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                            )
                            ** 0.5
                        )
                        + 1.0
                    )
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -(
                (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    -(
                        (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K2_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K3_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                    + 1.0
                )
            )
            + (
                K3_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K3_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K3_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
    )
    return result


######################
# SCOFe: O2, CO ######
######################


def SCOFe_SO_O2(xg_O2_, xg_S2_, constants):  # dmbCO_O
    """Evaluate differential of wtg_S - wtg_O with respect to xg_O2 for the S-C-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_S,
        wt_C,  # noqa
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        KD1_,
        KD2,
        y,
        y_S2_,
        y_O2_,
        y_SO2_,
        y_CO_,
        y_CO2_,
        y_OCS_,
        M_C,  # noqa
        M_S,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_S2,
        M_SO2,
        M_SO3,  # noqa
        M_OCS,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        + (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            -0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            * (
                -K1_
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * xg_O2_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K2_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K2_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                - M_CO2
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_CO2
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                - M_O2
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                K1_
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * xg_O2_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K2_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        ** 2
        - 0.111111111111111
        * (
            -2.0
            * K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - 3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        * (
            2.0
            * K4_
            * P
            * y_CO2_
            * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + 2.0
            * K4_
            * P
            * y_CO2_
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K3_ * y_OCS_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_CO_)
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            + 2.0
            * K4_
            * P
            * y_CO2_
            * (
                -K4_
                * M_CO2
                * P
                * y_CO2_
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - K4_
                * M_CO2
                * P
                * y_CO2_
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                - M_m_
                * (
                    -K4_
                    * P
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - K4_
                    * P
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + 4.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            * (
                -K1_
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * xg_O2_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K2_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K2_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                - M_CO2
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_CO2
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                - M_O2
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                K1_
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * xg_O2_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 2.0 * K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K2_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + 2.0
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                + 2.0
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - 0.5
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            + K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K3_ * y_OCS_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_CO_)
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            - 0.5
            * K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * xg_O2_
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        / (
            -0.666666666666667
            * K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + 0.333333333333333
            * (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            - 0.333333333333333
            * K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            -2.0
            * K4_
            * P
            * y_CO2_
            * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - 2.0
            * K4_
            * P
            * y_CO2_
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K3_ * y_OCS_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_CO_)
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            - 2.0
            * K4_
            * P
            * y_CO2_
            * (
                -K4_
                * M_CO2
                * P
                * y_CO2_
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - K4_
                * M_CO2
                * P
                * y_CO2_
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                - M_m_
                * (
                    -K4_
                    * P
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - K4_
                    * P
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - 4.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K3_ * y_OCS_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_CO_)
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            + 0.5
            * K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * xg_O2_
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        / (
            -2.0
            * K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - 3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
    )
    return result


def SCOFe_SO_S2(xg_O2_, xg_S2_, constants):  # dmbCO_CO
    """Evaluate differential of wtg_S - wtg_O with respect to xg_S2 for the S-C-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_S,
        wt_C,  # noqa
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        KD1_,
        KD2,
        y,
        y_S2_,
        y_O2_,
        y_SO2_,
        y_CO_,
        y_CO2_,
        y_OCS_,
        M_C,  # noqa
        M_S,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_S2,
        M_SO2,
        M_SO3,  # noqa
        M_OCS,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            -0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        + (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            * (
                0.5
                * K1_ ** (-4.0)
                * K2_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + 0.5
                * K1_ ** (-3.0)
                * K2_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - K1_
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -1.5
                    * K1_ ** (-4.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K2_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    1.0
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * xg_S2_ * y_OCS_)
                - M_CO2
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_S2
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                K1_
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -1.5
                    * K1_ ** (-4.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    1.0
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * xg_S2_ * y_OCS_)
                + 2.0
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        ** 2
        - 0.111111111111111
        * (
            -2.0
            * K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - 3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        * (
            -0.5
            * K1_ ** (-4.0)
            * K2_
            * K7_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CO
                * xg_S2_
                * y_OCS_
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            - 1.0
            * K1_ ** (-3.0)
            * K2_
            * K4_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_S2_
                * y_OCS_
                * (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            + 2.0
            * K4_
            * P
            * y_CO2_
            * (
                -0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + 2.0
            * K4_
            * P
            * y_CO2_
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * K4_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - K4_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_m_
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * K4_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - K4_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            * (
                0.5
                * K1_ ** (-4.0)
                * K2_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + 0.5
                * K1_ ** (-3.0)
                * K2_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - K1_
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -1.5
                    * K1_ ** (-4.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K2_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    1.0
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * xg_S2_ * y_OCS_)
                - M_CO2
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_S2
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                -0.5
                * K1_ ** (-4.0)
                * K2_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - 1.0
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + K1_
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -1.5
                    * K1_ ** (-4.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 1.0
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    1.0
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * xg_S2_ * y_OCS_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            + K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        / (
            -0.666666666666667
            * K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + 0.333333333333333
            * (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            - 0.333333333333333
            * K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            0.5
            * K1_ ** (-4.0)
            * K2_
            * K7_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CO
                * xg_S2_
                * y_OCS_
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            + 1.0
            * K1_ ** (-3.0)
            * K2_
            * K4_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_S2_
                * y_OCS_
                * (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            - 2.0
            * K4_
            * P
            * y_CO2_
            * (
                -0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - 2.0
            * K4_
            * P
            * y_CO2_
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * K4_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - K4_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_m_
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * K4_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - K4_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        / (
            -2.0
            * K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - 3.0
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + (
                2.0 * K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
    )
    return result


def SCOFe_SC_O2(xg_O2_, xg_S2_, constants):  # dmbCS_O
    """Evaluate differential of wtg_S - wtg_C with respect to xg_O2 for the S-C-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,  # noqa
        wt_S,
        wt_C,
        wt_Fe,  # noqa
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        KD1_,  # noqa
        KD2,  # noqa
        y,  # noqa
        y_S2_,
        y_O2_,
        y_SO2_,
        y_CO_,
        y_CO2_,
        y_OCS_,
        M_C,
        M_S,
        M_O,  # noqa
        M_Fe,  # noqa
        M_O2,
        M_CO,
        M_CO2,
        M_S2,
        M_SO2,
        M_SO3,  # noqa
        M_OCS,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        + (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            -0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            * (
                -K1_
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * xg_O2_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K2_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K2_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                - M_CO2
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_CO2
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                - M_O2
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                K1_
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * xg_O2_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K2_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        ** 2
        - (
            -K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + wt_C / M_C
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        * (
            K4_
            * P
            * y_CO2_
            * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + K4_
            * P
            * y_CO2_
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K3_ * y_OCS_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_CO_)
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            + K4_
            * P
            * y_CO2_
            * (
                -K4_
                * M_CO2
                * P
                * y_CO2_
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - K4_
                * M_CO2
                * P
                * y_CO2_
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                - M_m_
                * (
                    -K4_
                    * P
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - K4_
                    * P
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - (
                K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            * (
                -K1_
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * xg_O2_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K2_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K2_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                - M_CO2
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_CO2
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                - M_O2
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                K1_
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * xg_O2_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K2_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - 0.5
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            + K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K3_ * y_OCS_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_CO_)
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            - 0.5
            * K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * xg_O2_
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        / (
            -K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + (
                K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            -K4_
            * P
            * y_CO2_
            * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - K4_
            * P
            * y_CO2_
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K3_ * y_OCS_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_CO_)
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            - K4_
            * P
            * y_CO2_
            * (
                -K4_
                * M_CO2
                * P
                * y_CO2_
                * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - K4_
                * M_CO2
                * P
                * y_CO2_
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_CO_)
                )
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
                - M_m_
                * (
                    -K4_
                    * P
                    * y_CO2_
                    * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    - K4_
                    * P
                    * y_CO2_
                    * (
                        0.5
                        * K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_CO_)
                    )
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-K2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K3_ * y_OCS_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_CO_)
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            + 0.5
            * K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * xg_O2_
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        / (
            -K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + (
                K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
    )
    return result


def SCOFe_SC_S2(xg_O2_, xg_S2_, constants):  # dmbCS_CO
    """Evaluate differential of wtg_S - wtg_C with respect to xg_S2 for the S-C-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,  # noqa
        wt_S,
        wt_C,
        wt_Fe,  # noqa
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        KD1_,  # noqa
        KD2,  # noqa
        y,  # noqa
        y_S2_,
        y_O2_,
        y_SO2_,
        y_CO_,
        y_CO2_,
        y_OCS_,
        M_C,
        M_S,
        M_O,  # noqa
        M_Fe,  # noqa
        M_O2,
        M_CO,
        M_CO2,
        M_S2,
        M_SO2,
        M_SO3,  # noqa
        M_OCS,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            -0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        + (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5 * K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K6_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            * (
                0.5
                * K1_ ** (-4.0)
                * K2_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + 0.5
                * K1_ ** (-3.0)
                * K2_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - K1_
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -1.5
                    * K1_ ** (-4.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K2_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    1.0
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * xg_S2_ * y_OCS_)
                - M_CO2
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_S2
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                K1_
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -1.5
                    * K1_ ** (-4.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    1.0
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * xg_S2_ * y_OCS_)
                + 2.0
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        / (
            -K5_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K6_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + 2.0 * xg_S2_
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
        )
        ** 2
        - (
            -K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + wt_C / M_C
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        * (
            -0.5
            * K1_ ** (-4.0)
            * K2_
            * K7_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CO
                * xg_S2_
                * y_OCS_
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            - 0.5
            * K1_ ** (-3.0)
            * K2_
            * K4_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_S2_
                * y_OCS_
                * (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            + K4_
            * P
            * y_CO2_
            * (
                -0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + K4_
            * P
            * y_CO2_
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * K4_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - K4_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_m_
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * K4_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - K4_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - (
                K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            * (
                0.5
                * K1_ ** (-4.0)
                * K2_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + 0.5
                * K1_ ** (-3.0)
                * K2_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - K1_
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -1.5
                    * K1_ ** (-4.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K2_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    1.0
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * xg_S2_ * y_OCS_)
                - M_CO2
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_S2
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            ** 2
            - (
                -0.5
                * K1_ ** (-4.0)
                * K2_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - 0.5
                * K1_ ** (-3.0)
                * K2_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                + K1_
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    -1.5
                    * K1_ ** (-4.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                * (
                    1.0
                    * K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                / (
                    K3_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * xg_S2_ * y_OCS_)
                + (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            + K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        / (
            -K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + (
                K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            0.5
            * K1_ ** (-4.0)
            * K2_
            * K7_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CO
                * xg_S2_
                * y_OCS_
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            + 0.5
            * K1_ ** (-3.0)
            * K2_
            * K4_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_S2_
                * y_OCS_
                * (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                ** 2
            )
            - K4_
            * P
            * y_CO2_
            * (
                -0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - K4_
            * P
            * y_CO2_
            * (
                0.5
                * K1_ ** (-3.0)
                * K2_
                * K4_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_S2_
                    * y_OCS_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    ** 2
                )
                - K4_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K2_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                - M_m_
                * (
                    0.5
                    * K1_ ** (-3.0)
                    * K2_
                    * K4_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_S2_
                        * y_OCS_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        ** 2
                    )
                    - K4_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K2_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                ** 2
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
        / (
            -K4_
            * P
            * y_CO2_
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K4_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                    + M_m_
                    * (
                        -K4_
                        * P
                        * y_CO2_
                        * (
                            -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                        + 1.0
                    )
                )
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
            + (
                K2_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            / (
                K2_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K1_
                        * (
                            K1_ ** (-3.0)
                            * K2_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K3_ * y_OCS_)
                            + 1.0
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                        )
                    )
                )
                ** 3.0
                / (K3_ * y_OCS_)
                + M_CO2
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_CO_
                    * (
                        K1_ ** (-3.0)
                        * K2_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K3_ * y_OCS_)
                        + 1.0
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                    )
                )
            )
            - K7_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_CO
                * (
                    K1_ ** (-3.0)
                    * K2_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K3_ * y_OCS_)
                    + 1.0
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_CO_)
                )
            )
        )
    )
    return result


#####################
# CHOFe: O2, CO #####
#####################

# H2, CO, CH4 insoluble in the melt


def CHOFe_CO_O2(xg_O2_, xg_CO_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_O2 for the C-H-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,  # noqa
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        M_C,
        M_H,  # noqa
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -2.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 3.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.5
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + (0.5**0.5)
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (0.5**0.5)
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                (
                                    -4.0
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -2.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 3.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.5
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + (0.5**0.5)
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (0.5**0.5)
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                (
                                    -4.0
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - (
            -(
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -2.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 3.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.5
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                1.0
                * K2_
                * K5_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                + (0.5**0.5)
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (-2.0 * y + 1.0)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (-2.0 * y + 1.0)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -(
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - (
            -(
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            (
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -2.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 3.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.5
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + (
                1.0
                * K2_
                * K5_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                + (0.5**0.5)
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 2.0
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -(
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
    )
    return result


def CHOFe_CO_CO(xg_O2_, xg_CO_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_CO for the C-H-O-Fe system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        M_C,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -(
                2.0 * K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 2.0
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -(
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 2.0
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 2.0
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - (
            -(
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (-2.0 * y + 1.0)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (-2.0 * y + 1.0)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            (
                2.0 * K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 2.0
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 1
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        / (
            -(
                2.0 * K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
    )
    return result


def CHOFe_CH_O2(xg_O2_, xg_CO_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_O2 for the C-H-O-Fe system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        M_C,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (0.5**0.5)
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + (0.5**0.5)
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (0.5**0.5)
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                (
                                    -4.0
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + (0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -2.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 3.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.5
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.75
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -2.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 3.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.5
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + (0.5**0.5)
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (0.5**0.5)
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                (
                                    -4.0
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -2.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 3.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.5
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + (0.5**0.5)
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (0.5**0.5)
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                (
                                    -4.0
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - (
            -(0.5**0.5)
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 3.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + (0.5**0.5)
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (0.5**0.5)
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                (
                                    -4.0
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 3.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -2.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 3.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.5
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOFe_CH_CO(xg_O2_, xg_CO_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO for the C-H-O-Fe system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        M_C,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (0.5**0.5)
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + (0.5**0.5)
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (0.5**0.5)
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                (
                                    -4.0
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + (0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 2.0
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 2.0
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 2.0
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - (
            -(0.5**0.5)
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 2.0
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -4.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + (0.5**0.5)
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (0.5**0.5)
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                (
                                    -4.0
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - (0.5**0.5)
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -2.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 2.0
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -2.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 2.0
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + (0.5**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (0.5**0.5)
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            (
                                -4.0
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        (
                            -4.0
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    (
                        -4.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


# H2, CH4, CO SOLUBLE IN THE MELT


def CHOFe_CO_O2_2(xg_O2_, xg_CO_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_O2 for the C-H-O-Fe
    system with no dissolved H2, CH4, or CO in the melt.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,  # noqa
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,  # noqa
        K7_,
        K8_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        M_C,
        M_H,  # noqa
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + wt_C / M_C
        )
        * (
            0.25
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 0.375
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -0.25
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 0.375
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - 0.25
        * (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 2.0
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.5 * K7_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOFe_CO_CO_2(xg_O2_, xg_CO_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_CO for the C-H-O-Fe system with no dissolved H2, CH4, or CO in the melt.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        M_C,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + wt_C / M_C
        )
        * (
            0.25
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.25
            * K2_
            * K8_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K7_ * P * y_CO_ / M_CO
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -0.25
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * y_CO_ / M_CO
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - 0.25
        * (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K7_ * P * y_CO_ / M_CO
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 1
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.5 * K7_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K7_ * P * y_CO_ / M_CO
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOFe_CH_O2_2(xg_O2_, xg_CO_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_O2 for the C-H-O-Fe system with no dissolved H2, CH4, or CO in the melt.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        M_C,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + wt_C / M_C
        )
        * (
            0.25
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 0.375
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -0.5
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
        )
        * (
            0.5
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.75
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.75
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            + 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    - 0.125
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 1.0
                    / (K1_ * xg_O2_ * y_H2_)
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * xg_O2_ * y_CO_)
        )
        / (
            -0.5
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
        )
        ** 2
        - (
            -0.5
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.75
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    - 0.125
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 1.0
                    / (K1_ * xg_O2_ * y_H2_)
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * xg_O2_ * y_CO_)
        )
        / (
            -0.5
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
        )
        + (
            -0.25
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 0.375
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOFe_CH_CO_2(xg_O2_, xg_CO_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO for the C-H-O-Fe system with no dissolved H2, CH4, or CO in the melt.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        M_C,
        M_H,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + wt_C / M_C
        )
        * (
            0.25
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.25
            * K2_
            * K8_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K7_ * P * y_CO_ / M_CO
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -0.5
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
        )
        * (
            0.5
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K2_
            * K8_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            + 1.0
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    - 1.0
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_**2 * y_CO_)
        )
        / (
            -0.5
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
        )
        ** 2
        - (
            -0.5
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K2_
            * K8_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.0
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    - 1.0
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_**2 * y_CO_)
        )
        / (
            -0.5
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            - 0.5
            * K3_
            * K6_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
        )
        + (
            -0.25
            * K2_**2
            * K8_
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * y_CO_ / M_CO
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.25
            * K2_
            * K8_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K7_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


##############################
# SCHOFe: O2, CO, S2, H2 #####
##############################


def SCHOFe_OCS_CO_O2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_O2 for the S-C-H-O-Fe
    system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as
    guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,  # noqa
        wt_S,  # noqa
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,  # noqa
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,  # noqa
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,  # noqa
        M_H2S,
        M_OCS,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 0.5**0.5
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            + 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 2.0 * K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -(2 / 3)
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.235702260395516
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 0.5**0.5
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCS_CO_CO(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_CO for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5**0.5
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 1
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -(2 / 3)
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.235702260395516
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.5**0.5
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCS_CO_S2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_S2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K2_
        * K5_
        * P
        * xg_CO_
        * y_CO_
        * (P * xg_O2_ * y_O2_) ** 0.5
        * (
            -(0.125**0.5)
            * M_H2O
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            + 0.125**0.5
            * M_m_
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
        )
        / (
            (
                -K2_
                * K5_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + xg_CO_
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / y_SO2_
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_S2 * xg_S2_
                    + 0.5
                    * K3_
                    * K7_
                    * M_H2S
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 1.5
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            + 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.125**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -(2 / 3)
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.235702260395516
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.125**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
        )
        / (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCS_CH_O2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_O2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            0.5**0.5
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.75
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - (
            -(0.5**0.5)
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCS_CH_CO(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            0.5**0.5
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1.0
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - (
            -(0.5**0.5)
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCS_CH_S2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_S2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K2_
        * K5_
        * P
        * xg_CO_
        * y_CO_
        * (P * xg_O2_ * y_O2_) ** 0.5
        * (
            -(0.125**0.5)
            * M_H2O
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            + 0.125**0.5
            * M_m_
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
        )
        / (
            (
                -K2_
                * K5_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + xg_CO_
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / y_SO2_
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_S2 * xg_S2_
                    + 0.5
                    * K3_
                    * K7_
                    * M_H2S
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 1.5
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        - (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            0.5**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.125**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.25
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.125**0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
        )
        / (
            -(0.5**0.5)
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCS_CS_O2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_O2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            -0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCS_CS_CO(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_CO for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -(
                1.0
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        * (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.5**0.5
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCS_CS_S2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_S2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K2_
        * K5_
        * P
        * xg_CO_
        * y_CO_
        * (P * xg_O2_ * y_O2_) ** 0.5
        * (
            -(0.125**0.5)
            * M_H2O
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            + 0.125**0.5
            * M_m_
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
        )
        / (
            (
                -K2_
                * K5_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.5**0.5
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.5**0.5
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + xg_CO_
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / y_SO2_
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_S2 * xg_S2_
                    + 0.5
                    * K3_
                    * K7_
                    * M_H2S
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 1.5
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        - (
            -0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -(0.125**0.5)
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.125**0.5
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.5**0.5
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.5**0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 2.0
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.25
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe_OHS_HO_O2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_O with respect to xg_S2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0 * K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                + 2.0
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 0.5
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -(2 / 3)
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - (1 / 3)
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OHS_HO_H2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_O with respect to xg_H2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 1
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 4.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 4.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 4.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 4.0
            * K5_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_H2_
                * y_CH4_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -(2 / 3)
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - (1 / 3)
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 4.0
            * K5_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_H2_
                * y_CH4_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OHS_HO_S2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_O with respect to xg_S2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
        * (
            -K5_
            * M_CO2
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                1.0
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                + K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            - M_m_
            * (
                -K5_
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            + 0.5
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * M_CO2
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            (
                -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
                / (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + xg_H2_
                    + 2.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                / (
                    K1_
                    * K7_
                    * M_H2S
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_
                    * M_H2O
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2O_
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / y_SO2_
                    + M_CO2
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_S2 * xg_S2_
                    + M_CH4
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + M_CO
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * y_CO_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    ** (-2.0)
                    * (
                        y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K2_
                            * (
                                1.0
                                + y_CO2_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                + y_CO2_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K2_ * y_CO_)
                                + K2_ ** (-3.0)
                                * K6_
                                * P ** (-0.5)
                                * xg_O2_
                                * y_CO2_
                                * y_O2_
                                * (xg_O2_ * y_O2_) ** (-1.5)
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / (K10_ * y_OCS_)
                            )
                        )
                    )
                    ** 3.0
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 1.0
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 0.5
                * K2_ ** (-4.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 1.0
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - 1.0
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -(2 / 3)
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - (1 / 3)
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OHS_HC_O2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_C with respect to xg_O2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + wt_C / M_C
        )
        * (
            K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K5_
            * P
            * y_CO2_
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 0.5
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
        )
        / (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OHS_HC_H2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_C with respect to xg_H2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 1
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 4.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 4.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + wt_C / M_C
        )
        * (
            K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 2.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - 2.0
            * K5_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_H2_
                * y_CH4_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_H2_
                * y_CH4_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OHS_HC_S2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_C with respect to xg_S2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
        * (
            -K5_
            * M_CO2
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                1.0
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                + K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            - M_m_
            * (
                -K5_
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            + 0.5
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * M_CO2
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            (
                -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
                / (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + xg_H2_
                    + 2.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                / (
                    K1_
                    * K7_
                    * M_H2S
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_
                    * M_H2O
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2O_
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / y_SO2_
                    + M_CO2
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_S2 * xg_S2_
                    + M_CH4
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + M_CO
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * y_CO_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    ** (-2.0)
                    * (
                        y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K2_
                            * (
                                1.0
                                + y_CO2_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                + y_CO2_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K2_ * y_CO_)
                                + K2_ ** (-3.0)
                                * K6_
                                * P ** (-0.5)
                                * xg_O2_
                                * y_CO2_
                                * y_O2_
                                * (xg_O2_ * y_O2_) ** (-1.5)
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / (K10_ * y_OCS_)
                            )
                        )
                    )
                    ** 3.0
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 1.0
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + wt_C / M_C
        )
        * (
            K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 0.5
                * K2_ ** (-4.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 0.5
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OHS_HS_O2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_S with respect to xg_O2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        - (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            -0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OHS_HS_H2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_S with respect to xg_H2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 1
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 4.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 4.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -(
                K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        * (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OHS_HS_S2(xg_O2_, xg_H2_, xg_S2_, constants):
    """Evaluate differential of wtg_H - wtg_S with respect to xg_S2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
        * (
            -K5_
            * M_CO2
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                1.0
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                + K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            - M_m_
            * (
                -K5_
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            + 0.5
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * M_CO2
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            (
                -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
                / (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + xg_H2_
                    + 2.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                / (
                    K1_
                    * K7_
                    * M_H2S
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_
                    * M_H2O
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2O_
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / y_SO2_
                    + M_CO2
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_S2 * xg_S2_
                    + M_CH4
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + M_CO
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * y_CO_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    ** (-2.0)
                    * (
                        y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K2_
                            * (
                                1.0
                                + y_CO2_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                + y_CO2_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K2_ * y_CO_)
                                + K2_ ** (-3.0)
                                * K6_
                                * P ** (-0.5)
                                * xg_O2_
                                * y_CO2_
                                * y_O2_
                                * (xg_O2_ * y_O2_) ** (-1.5)
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / (K10_ * y_OCS_)
                            )
                        )
                    )
                    ** 3.0
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        + (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 1.0
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        - (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 2.0
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe_OCH_CO_O2(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_O2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / (M_S * xg_O2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 2.0
                * K6_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -(2 / 3)
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (1 / 3)
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / (M_S * xg_O2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCH_CO_CO(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_CO for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 1
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -(2 / 3)
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (1 / 3)
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCH_CO_H2(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_H2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K2_
        * K5_
        * P
        * xg_CO_
        * y_CO_
        * (P * xg_O2_ * y_O2_) ** 0.5
        * (
            -0.5
            * M_H2O
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / xg_H2_
            + 0.5
            * M_m_
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / xg_H2_
        )
        / (
            (
                -K2_
                * K5_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + xg_CO_
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (
                    K1_
                    * K7_
                    * M_H2S
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_
                    * M_H2O
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2O_
                    + K2_
                    * M_CO2
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_CO2_
                    + K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    / y_SO2_
                    + M_CO * xg_CO_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_S2
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (1 / 9)
        * (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -(2 / 3)
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (1 / 3)
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (1 / 3)
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCH_CH_O2(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_O2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + 0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCH_CH_CO(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
        * (
            -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
        )
        / (
            (
                -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
                / (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + 2.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + xg_H2_
                )
                / (
                    K1_
                    * K7_
                    * M_H2S
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_
                    * M_H2O
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2O_
                    + K2_
                    * M_CO2
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_CO2_
                    + K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    / y_SO2_
                    + M_CO * xg_CO_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_S2
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        - (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + 2.0
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCH_CH_H2(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_H2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K2_
        * K5_
        * P
        * xg_CO_
        * y_CO_
        * (P * xg_O2_ * y_O2_) ** 0.5
        * (
            -0.5
            * M_H2O
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / xg_H2_
            + 0.5
            * M_m_
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / xg_H2_
        )
        / (
            (
                -K2_
                * K5_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + xg_CO_
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (
                    K1_
                    * K7_
                    * M_H2S
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_
                    * M_H2O
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2O_
                    + K2_
                    * M_CO2
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_CO2_
                    + K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    / y_SO2_
                    + M_CO * xg_CO_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_S2
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        - (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 4.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                + 1
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCH_CS_O2(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_O2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + wt_S / M_S
        )
        * (
            K8_
            * xg_O2_
            * y_O2_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                0.5
                * y_S2_
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                / (xg_O2_ * y_O2_)
                - 0.5
                * y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_**2 * y_O2_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-2.0)
            / (M_S * y_S2_)
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / (M_S * xg_O2_)
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + K6_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K8_
            * xg_O2_
            * y_O2_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                0.5
                * y_S2_
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                / (xg_O2_ * y_O2_)
                - 0.5
                * y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_**2 * y_O2_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-2.0)
            / (M_S * y_S2_)
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCH_CS_CO(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_CO for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -(
            -0.5
            * K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
        )
        / (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 2.0
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe_OCH_CS_H2(xg_O2_, xg_CO_, xg_H2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_H2 for the S-C-H-O-Fe system with no dissolved H2, CO, CH4, or H2S in the melt and O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K2_
        * K5_
        * P
        * xg_CO_
        * y_CO_
        * (P * xg_O2_ * y_O2_) ** 0.5
        * (
            -0.5
            * M_H2O
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / xg_H2_
            + 0.5
            * M_m_
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / xg_H2_
        )
        / (
            (
                -K2_
                * K5_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                + (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + xg_CO_
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (
                    K1_
                    * K7_
                    * M_H2S
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_
                    * M_H2O
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2O_
                    + K2_
                    * M_CO2
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_CO2_
                    + K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    / y_SO2_
                    + M_CO * xg_CO_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_S2
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (
                        P
                        * y_S2_
                        * (
                            -0.5
                            * K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + 1.0
                            * (
                                -K1_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_H2O_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                - K2_
                                * xg_CO_
                                * y_CO_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                - xg_CO_
                                - xg_H2_
                                - xg_O2_
                                + 0.25
                                * (
                                    K1_
                                    * K7_
                                    * xg_H2_
                                    * y_H2_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                    / y_H2S_
                                    + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                    + K6_
                                    * P**1.5
                                    * xg_O2_
                                    * y_O2_
                                    * y_S2_**0.5
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                ** 2.0
                                + 1
                            )
                            ** 0.5
                            - 0.5
                            * K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
            )
            * (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        - (
            -0.5
            * K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
        )
        / (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 2.0
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


###########################
# SCHOFe2: O2, CO, S2 #####
###########################


def SCHOFe2_OCS_CO_O2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_O2 for the S-C-H-O-Fe
    system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,  # noqa
        wt_S,  # noqa
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,  # noqa
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,  # noqa
        K14_,  # noqa
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,  # noqa
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,  # noqa
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,  # noqa
        M_H2S,
        M_OCS,
        M_FeO,  # noqa
        M_FeO15,  # noqa
        M_m_,
        M_m_ox,  # noqa
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 2.0 * K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.235702260395516
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCS_CO_CO(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_CO for the S-C-H-O-Fe system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K11_ * P * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 1
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.235702260395516
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCS_CO_S2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_O with respect to xg_S2 for the S-C-H-O-Fe system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.235702260395516
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe2_OCS_CH_O2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_O2 for the S-C-H-O-Fe system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.75
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * xg_O2_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_O2_ * y_CO_)
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.75
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.75
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * xg_O2_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_O2_ * y_CO_)
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCS_CH_CO(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO for the S-C-H-O-Fe system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + 1.0
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_**2 * y_CO_)
            + 1.0
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_**2 * y_CO_)
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1.0
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 1.0
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_**2 * y_CO_)
            - 1.0
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_**2 * y_CO_)
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCS_CH_S2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_S2 for the S-C-H-O-Fe system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            1.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.25
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_S2_ * y_CO_)
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.25
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -1.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.25
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_S2_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCS_CS_O2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_O2 for the S-C-H-O-Fe system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_O2_ * y_CO_)
            + 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_O2_ * y_CO_)
            - 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            0.5
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                + 0.125
                                * (
                                    -1.0
                                    * K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / (xg_O2_ * y_H2S_)
                                    - 1.0
                                    * y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * xg_O2_ * y_H2_)
                                )
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                -K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + K6_
                                        * xg_O2_
                                        * y_O2_
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        / y_SO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_S2_
                                        - 1.0
                                        + K6_
                                        * P
                                        * xg_O2_
                                        * y_O2_
                                        * (xg_CO_ * y_CO_) ** 3.0
                                        * (P * xg_S2_ * y_S2_) ** 0.5
                                        * (
                                            K2_
                                            * xg_CO_
                                            * y_CO_
                                            * (P * xg_O2_ * y_O2_) ** 0.5
                                        )
                                        ** (-2.0)
                                        / (K10_ * y_OCS_)
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        K7_
                                        * y_H2O_
                                        * (xg_O2_ * y_O2_) ** (-0.5)
                                        * (xg_S2_ * y_S2_) ** 0.5
                                        / y_H2S_
                                        + 1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCS_CS_CO(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_CO for the S-C-H-O-Fe system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        -(
            -1.0
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_**2 * y_CO_)
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            1.0
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_**2 * y_CO_)
            - (
                1.0
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                                + 1
                                + 1.0
                                * K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * xg_CO_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCS_CS_S2(xg_O2_, xg_CO_, xg_S2_, constants):
    """Evaluate differential of wtg_C - wtg_S with respect to xg_S2 for the S-C-H-O-Fe system with O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.353553390593274
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.353553390593274
                * M_m_
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.25
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_S2_ * y_CO_)
            - 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.25
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_S2_ * y_CO_)
            + 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 2.0
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.25
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


####################
# SCHOFe2: OHS #####
####################


def SCHOFe2_OHS_HO_O2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHO_O
    """Evaluate differential of wtg_H - wtg_O with respect to xg_S2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + 0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - 0.111111111111111
        * (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 0.5
            * K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0 * K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                + 2.0
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 0.5
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -0.333333333333333
            * K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.666666666666667
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.333333333333333
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 0.5
            * K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OHS_HO_H2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHO_H
    """Evaluate differential of wtg_H - wtg_O with respect to xg_H2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 4.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_H2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 4.0
            * K12_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_**2
                * M_CH4
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + K13_ * P * y_H2_ / M_H2
            + K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 1
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 4.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 4.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 4.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_H2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 4.0
            * K12_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_**2
                * M_CH4
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - K13_ * P * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        - 0.111111111111111
        * (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K11_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * K3_
                * M_CO
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 4.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 4.0
            * K5_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_H2_
                * y_CH4_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -0.333333333333333
            * K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.666666666666667
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.333333333333333
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K11_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * K3_
                * M_CO
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 4.0
            * K5_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_H2_
                * y_CH4_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OHS_HO_S2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHO_S
    """Evaluate differential of wtg_H - wtg_O with respect to xg_S2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (M_H2S * xg_S2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K12_
            * K2_ ** (-3.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-3.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * K3_
                * M_CH4
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + 0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (M_H2S * xg_S2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 1.0
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - 1.0
            * K12_
            * K2_ ** (-3.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-3.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * K3_
                * M_CH4
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - 0.111111111111111
        * (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 0.5
                * K2_ ** (-4.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 1.0
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K11_
            * K2_ ** (-4.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * M_CO
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 1.0
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -0.333333333333333
            * K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.666666666666667
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.333333333333333
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K11_
            * K2_ ** (-4.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * M_CO
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 1.0
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OHS_HC_O2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHC_O
    """Evaluate differential of wtg_H - wtg_C with respect to xg_O2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        -(
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + wt_C / M_C
        )
        * (
            K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 0.5
            * K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 1.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K5_
            * P
            * y_CO2_
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 0.5
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + 0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 0.5
            * K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 1.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OHS_HC_H2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHC_H
    """Evaluate differential of wtg_H - wtg_C with respect to xg_H2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        -(
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + wt_C / M_C
        )
        * (
            K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K11_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * K3_
                * M_CO
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_H2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K12_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_**2
                * M_CH4
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 2.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - 2.0
            * K5_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_H2_
                * y_CH4_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + 0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 4.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_H2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 4.0
            * K12_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_**2
                * M_CH4
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + K13_ * P * y_H2_ / M_H2
            + K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 1
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 4.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 4.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 4.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_H2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 4.0
            * K12_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_**2
                * M_CH4
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - K13_ * P * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        - (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K11_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * K3_
                * M_CO
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_H2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K12_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_**2
                * M_CH4
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K5_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * xg_H2_
                * y_CH4_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OHS_HC_S2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHC_S
    """Evaluate differential of wtg_H - wtg_C with respect to xg_S2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        -(
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + wt_C / M_C
        )
        * (
            K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 0.5
                * K2_ ** (-4.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K11_
            * K2_ ** (-4.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * M_CO
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 0.5
            * K12_
            * K2_ ** (-3.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-3.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * K3_
                * M_CH4
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 0.5
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (M_H2S * xg_S2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K12_
            * K2_ ** (-3.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-3.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * K3_
                * M_CH4
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + 0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (M_H2S * xg_S2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 1.0
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - 1.0
            * K12_
            * K2_ ** (-3.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-3.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * K3_
                * M_CH4
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                ** 2
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 0.5
            * K11_
            * K2_ ** (-4.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * M_CO
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 0.5
            * K12_
            * K2_ ** (-3.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-3.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * K3_
                * M_CH4
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 0.5
            * K2_ ** (-3.0)
            * K5_
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-1.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * xg_S2_
                * y_OCS_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K11_
            * P
            * y_CO2_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K2_
                * M_CO
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K5_
            * P
            * y_CO2_
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (
                (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OHS_HS_O2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHS_O
    """Evaluate differential of wtg_H - wtg_S with respect to xg_O2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        -(
            0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        - (
            -0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        * (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + 0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                1.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5 * y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * xg_O2_ * y_CO_)
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - 2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_O2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - K5_
                    * P
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_CO2
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
                - M_O2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 1.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_O2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    1.0
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K2_ * xg_O2_ * y_CO_)
                    + 0.5
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * xg_O2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + 3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                    - 1.5
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * xg_O2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 2.0
                    * y_CO2_
                    * (
                        1.0
                        * y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.5
                        * y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K2_ * xg_O2_ * y_CO_)
                        + 0.5
                        * K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe2_OHS_HS_H2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHS_H
    """Evaluate differential of wtg_H - wtg_S with respect to xg_O2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        K14_
        * K1_
        * K7_
        * P
        * y_H2_
        * (xg_O2_ * y_O2_) ** (-0.5)
        * (xg_S2_ * y_S2_) ** 0.5
        * (P * xg_O2_ * y_O2_) ** 0.5
        / (
            M_H2S
            * (
                -K14_
                * K1_
                * K7_
                * P
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / M_H2S
                - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
                - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
                + (
                    K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + 2.0 * xg_S2_
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    ** (-2.0)
                    * (
                        y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K2_
                            * (
                                1.0
                                + y_CO2_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                + y_CO2_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K2_ * y_CO_)
                                + K2_ ** (-3.0)
                                * K6_
                                * P ** (-0.5)
                                * xg_O2_
                                * y_CO2_
                                * y_O2_
                                * (xg_O2_ * y_O2_) ** (-1.5)
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / (K10_ * y_OCS_)
                            )
                        )
                    )
                    ** 3.0
                    / (K10_ * y_OCS_)
                )
                / (
                    K1_
                    * K7_
                    * M_H2S
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    + K1_
                    * M_H2O
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2O_
                    + K6_
                    * M_SO2
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / y_SO2_
                    + M_CO2
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_S2 * xg_S2_
                    + M_CH4
                    * y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + M_CO
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * y_CO_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    + K6_
                    * M_OCS
                    * P
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    ** (-2.0)
                    * (
                        y_CO2_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K2_
                            * (
                                1.0
                                + y_CO2_
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** 2.0
                                / (K3_ * y_CH4_)
                                + y_CO2_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K2_ * y_CO_)
                                + K2_ ** (-3.0)
                                * K6_
                                * P ** (-0.5)
                                * xg_O2_
                                * y_CO2_
                                * y_O2_
                                * (xg_O2_ * y_O2_) ** (-1.5)
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / (K10_ * y_OCS_)
                            )
                        )
                    )
                    ** 3.0
                    / (K10_ * y_OCS_)
                )
            )
        )
        - (
            K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (
                K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        * (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + 0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 4.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_H2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 4.0
            * K12_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_**2
                * M_CH4
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            + K13_ * P * y_H2_ / M_H2
            + K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_CO2
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_H2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 2.0
                * M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO2
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 2.0
                * M_CH4
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 2.0
                * M_CO
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * K3_
                    * xg_H2_
                    * y_CH4_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 6.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 4.0
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 1
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 4.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 4.0
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_**2
                    * xg_H2_
                    * y_CH4_**2
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 4.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * xg_H2_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 4.0
            * K12_
            * P
            * y_CO2_**2
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 4.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_**2
                * M_CH4
                * xg_H2_
                * y_CH4_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
            - K13_ * P * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - 0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - 0.5
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_H2_
                    + 2.0
                    * K5_
                    * P
                    * y_CO2_**2
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K3_
                        * xg_H2_
                        * y_CH4_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 2.0
                * K5_
                * M_CO2
                * P
                * y_CO2_**2
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * xg_H2_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K5_
                    * M_CO2
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K5_
                        * P
                        * y_CO2_
                        * (
                            -K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2S_
                            - K1_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_H2O_
                            - K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            - xg_H2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OHS_HS_S2(xg_O2_, xg_H2_, xg_S2_, constants):  # dmbHS_S
    """Evaluate differential of wtg_H - wtg_S with respect to xg_S2 for the S-C-H-O-Fe
    system with O2, H2, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        -(
            -0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (M_H2S * xg_S2_)
            - 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        + (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (M_H2S * xg_S2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K12_
            * K2_ ** (-3.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-3.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * K3_
                * M_CH4
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        - (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (M_H2S * xg_S2_)
            + 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 2.0
                + K2_
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + 0.25
        * (
            -2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            2.0
            * K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            + 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (M_H2S * xg_S2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 0.5
                    * K2_ ** (-3.0)
                    * K5_
                    * K6_
                    * P**0.5
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                + 0.5
                * K2_ ** (-3.0)
                * K5_
                * K6_
                * M_CO2
                * P**0.5
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - 1.0
                * K2_ ** (-3.0)
                * K6_
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_CO2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                - M_S2
                - M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                - M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    - 0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + 0.5
                * K2_ ** (-4.0)
                * K6_
                * M_CO
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_CO_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CO2
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-1.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * xg_S2_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                + 0.5
                * K2_ ** (-3.0)
                * K6_
                * M_CH4
                * P ** (-0.5)
                * xg_O2_
                * y_CO2_**2
                * y_O2_
                * (xg_O2_ * y_O2_) ** (-3.5)
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K10_
                    * K3_
                    * xg_S2_
                    * y_CH4_
                    * y_OCS_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    ** 2
                )
                - K2_
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    3.0
                    * y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                    - 1.5
                    * K2_ ** (-4.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                * (
                    -2.0
                    * y_CO2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_S2_ * y_H2S_)
                        - 0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    + 1.0
                    * K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_**2
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K10_
                        * xg_S2_
                        * y_OCS_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                        ** 2
                    )
                )
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                / (
                    K10_
                    * y_CO2_
                    * y_OCS_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
            ** 2
            - 1.0
            * K12_
            * K2_ ** (-3.0)
            * K6_
            * P**0.5
            * xg_O2_
            * y_CO2_**2
            * y_O2_
            * (xg_O2_ * y_O2_) ** (-3.5)
            * (P * xg_S2_ * y_S2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K10_
                * K3_
                * M_CH4
                * xg_S2_
                * y_OCS_
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                ** 2
            )
        )
        / (
            -K12_
            * P
            * y_CO2_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            * (
                -K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_H2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K3_
                * M_CH4
                * (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
            )
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K5_
                * M_CO2
                * P
                * y_CO2_
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K5_
                    * P
                    * y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + xg_H2_
                + 2.0
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO2
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    1.0
                    + y_CO2_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                    + K2_ ** (-3.0)
                    * K6_
                    * P ** (-0.5)
                    * xg_O2_
                    * y_CO2_
                    * y_O2_
                    * (xg_O2_ * y_O2_) ** (-1.5)
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (K10_ * y_OCS_)
                )
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_CH4
                * y_CO2_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + M_CO
                * y_CO2_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / y_H2S_
                    - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_H2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K2_
                    * y_CO_
                    * (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (
                    y_CO2_
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        1.0
                        + y_CO2_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                        + K2_ ** (-3.0)
                        * K6_
                        * P ** (-0.5)
                        * xg_O2_
                        * y_CO2_
                        * y_O2_
                        * (xg_O2_ * y_O2_) ** (-1.5)
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (K10_ * y_OCS_)
                    )
                )
                ** (-2.0)
                * (
                    y_CO2_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / y_H2S_
                        - K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_H2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K2_
                        * (
                            1.0
                            + y_CO2_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            + y_CO2_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K2_ * y_CO_)
                            + K2_ ** (-3.0)
                            * K6_
                            * P ** (-0.5)
                            * xg_O2_
                            * y_CO2_
                            * y_O2_
                            * (xg_O2_ * y_O2_) ** (-1.5)
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (K10_ * y_OCS_)
                        )
                    )
                )
                ** 3.0
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


####################
# SCHOFe2: OCH #####
####################


def SCHOFe2_OCH_CO_O2(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCO_O
    """Evaluate differential of wtg_C - wtg_O with respect to xg_S2 for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / (M_S * xg_O2_)
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 2.0
                * K6_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - 0.333333333333333
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / (M_S * xg_O2_)
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCH_CO_CO(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCO_CO
    """Evaluate differential of wtg_C - wtg_O with respect to xg_CO for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K11_ * P * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 1
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - 0.333333333333333
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe2_OCH_CO_H2(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCO_H
    """Evaluate differential of wtg_C - wtg_O with respect to xg_H2 for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_H2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_H2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            - 0.333333333333333
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.333333333333333
            * (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe2_OCH_CH_O2(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCH_O
    """Evaluate differential of wtg_C - wtg_H with respect to xg_O2 for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            1.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - 0.25
        * (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            -1.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + 0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.25
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe2_OCH_CH_CO(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCH_CO
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        -(
            -2.0
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - 0.25
        * (
            2.0
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            + 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + 2.0
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        * (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        / (
            -K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe2_OCH_CH_H2(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCH_H
    """Evaluate differential of wtg_C - wtg_H with respect to xg_H2 for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_H2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_H2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - 0.25
        * (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            4.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_H2_)
            + K13_ * P * y_H2_ / M_H2
            + 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            + K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            + (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 4.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                + 1
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            + 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - 0.5 * K13_ * P * xg_H2_ * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -4.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_H2_)
            - K13_ * P * y_H2_ / M_H2
            - 0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            - K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K13_ * P * xg_H2_ * y_H2_ / M_H2
            - K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_H2_
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCH_CS_O2(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCS_O
    """Evaluate differential of wtg_C - wtg_S with respect to xg_O2 for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.25
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.25
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    / xg_O2_
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_H2O
                    * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            + K8_
            * xg_O2_
            * y_O2_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                0.5
                * y_S2_
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                / (xg_O2_ * y_O2_)
                - 0.5
                * y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_**2 * y_O2_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-2.0)
            / (M_S * y_S2_)
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / (M_S * xg_O2_)
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - 0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - K6_
                * M_SO2
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                - M_O2
                - M_S2
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + K6_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            - K8_
            * xg_O2_
            * y_O2_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                0.5
                * y_S2_
                * (
                    -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 2.0
                    * (
                        -0.25
                        * K1_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_H2O_)
                        - 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 0.25
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.25
                        * K6_
                        * y_O2_
                        * (P * y_S2_) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_SO2_
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                / (xg_O2_ * y_O2_)
                - 0.5
                * y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_**2 * y_O2_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-2.0)
            / (M_S * y_S2_)
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0 * K6_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                * (
                    -0.25
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    + 0.25
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.25
                    * K6_
                    * y_O2_
                    * (P * y_S2_) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_SO2_
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


def SCHOFe2_OCH_CS_CO(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCS_CO
    """Evaluate differential of wtg_C - wtg_S with respect to xg_CO for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        -(
            -0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            - 0.5
            * K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K2_ * K5_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            + 0.5
            * K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                2.0
                * (
                    -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - 0.5
                    * K2_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - 0.5
                    + 0.25
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
                - 1.0
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_CO
                - M_S2
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 2.0
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    2.0
                    * (
                        -0.5 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - 0.5
                        * K2_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - 0.5
                        + 0.25
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                    - 1.0
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return result


def SCHOFe2_OCH_CS_H2(xg_O2_, xg_CO_, xg_H2_, constants):  # dmbCS_H
    """Evaluate differential of wtg_C - wtg_S with respect to xg_H2 for the S-C-H-O-Fe
    system with O2, CO, and H2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_H2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            2.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4 * xg_H2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_H2O
                * (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - (K1_ * K4_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            + K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            + 0.5
            * K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                - K1_
                * K7_
                * M_H2S
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                - K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - 2.0
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * xg_H2_ * y_CH4_)
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                - M_H2
                - M_S2
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_H2S_
                + K1_
                * K7_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + 0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / y_SO2_
                + 2.0
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 1.0
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                * (
                    -1.0
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        0.25
                        * K1_
                        * K7_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 1.0
                        / y_H2S_
                        - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - 1.0
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * xg_H2_ * y_CH4_)
                        - 0.5
                    )
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** (-0.5)
                )
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** (-1.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_H2S
            - K14_
            * K1_
            * K7_
            * P
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - 0.5
            * K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (
                -1.0
                * K1_
                * K7_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    0.25
                    * K1_
                    * K7_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 1.0
                    / y_H2S_
                    - 0.5 * K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - 1.0
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * xg_H2_ * y_CH4_)
                    - 0.5
                )
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** (-0.5)
            )
            * (
                -0.5
                * K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                / y_H2S_
                - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                + 1.0
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - K2_
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                    / (K3_ * y_CH4_)
                    - xg_CO_
                    - xg_H2_
                    - xg_O2_
                    + 0.25
                    * (
                        K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                    + 1
                )
                ** 0.5
                - 0.5
                * K6_
                * P**1.5
                * xg_O2_
                * y_O2_
                * y_S2_**0.5
                * (xg_CO_ * y_CO_) ** 3.0
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** (-1.0)
            / M_S
        )
        / (
            -K14_
            * K1_
            * K7_
            * P
            * xg_H2_
            * y_H2_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            * (P * xg_O2_ * y_O2_) ** 0.5
            / M_H2S
            - K8_
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                / (xg_O2_ * y_O2_)
            )
            ** 0.5
            / M_S
            - K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (
                y_S2_
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
            )
            ** 0.5
            / M_S
            + (
                K1_
                * K7_
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K6_
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + 2.0
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K1_
                * K7_
                * M_H2S
                * xg_H2_
                * y_H2_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (
                    y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (P * xg_O2_ * y_O2_) ** 0.5
                / y_H2S_
                + K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                / (K3_ * y_CH4_)
                + K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                / y_SO2_
                + M_CO * xg_CO_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_S2
                * (
                    -0.5
                    * K1_
                    * K7_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                    / y_H2S_
                    - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                    + 1.0
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - K2_
                        * xg_CO_
                        * y_CO_
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 2.0
                        / (K3_ * y_CH4_)
                        - xg_CO_
                        - xg_H2_
                        - xg_O2_
                        + 0.25
                        * (
                            K1_
                            * K7_
                            * xg_H2_
                            * y_H2_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                            / y_H2S_
                            + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                            + K6_
                            * P**1.5
                            * xg_O2_
                            * y_O2_
                            * y_S2_**0.5
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        ** 2.0
                        + 1
                    )
                    ** 0.5
                    - 0.5
                    * K6_
                    * P**1.5
                    * xg_O2_
                    * y_O2_
                    * y_S2_**0.5
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                ** 2.0
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (
                    P
                    * y_S2_
                    * (
                        -0.5
                        * K1_
                        * K7_
                        * xg_H2_
                        * y_H2_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                        / y_H2S_
                        - 0.5 * K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                        + 1.0
                        * (
                            -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / y_CO2_
                            - K2_
                            * xg_CO_
                            * y_CO_
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** 2.0
                            / (K3_ * y_CH4_)
                            - xg_CO_
                            - xg_H2_
                            - xg_O2_
                            + 0.25
                            * (
                                K1_
                                * K7_
                                * xg_H2_
                                * y_H2_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (y_S2_ / (xg_O2_ * y_O2_)) ** 0.5
                                / y_H2S_
                                + K6_ * xg_O2_ * y_O2_ * (P * y_S2_) ** 0.5 / y_SO2_
                                + K6_
                                * P**1.5
                                * xg_O2_
                                * y_O2_
                                * y_S2_**0.5
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            ** 2.0
                            + 1
                        )
                        ** 0.5
                        - 0.5
                        * K6_
                        * P**1.5
                        * xg_O2_
                        * y_O2_
                        * y_S2_**0.5
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    ** 2.0
                )
                ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return result


###############
# CH: CO2 #####
###############


def CH_CO2(xg_CO2_, values):
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO2 for the C-H-O-Fe
    system with only H2O and CO2.

    Args:
        xg_CO2_ (float): Mole fraction of CO2 in the gas
        values (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    P, wt_C, wt_H, K4_, K5_, y_H2O_, y_CO2_, M_H, M_C, M_CO2, M_H2O, M_m_ = values
    result = (
        -(
            -((K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5)
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
            * (
                -K5_ * M_CO2 * P * y_CO2_
                + 0.5
                * M_H2O
                * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                / (1.0 - xg_CO2_)
                - M_m_
                * (
                    -K5_ * P * y_CO2_
                    + 0.5
                    * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    / (1.0 - xg_CO2_)
                )
            )
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
            / (
                (1.0 - xg_CO2_)
                * (
                    K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                    + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + M_m_
                    * (
                        -K5_ * P * xg_CO2_ * y_CO2_
                        - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                        + 1.0
                    )
                )
            )
            - (1.0 - xg_CO2_)
            * (-M_CO2 + M_H2O)
            / (M_CO2 * xg_CO2_ + M_H2O * (1.0 - xg_CO2_)) ** 2
            + 1 / (M_CO2 * xg_CO2_ + M_H2O * (1.0 - xg_CO2_))
        )
        / (
            -((K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5)
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            + (1.0 - xg_CO2_) / (M_CO2 * xg_CO2_ + M_H2O * (1.0 - xg_CO2_))
        )
        ** 2
        + (
            -K5_
            * P
            * xg_CO2_
            * y_CO2_
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K5_
            * P
            * xg_CO2_
            * y_CO2_
            * (
                -K5_ * M_CO2 * P * y_CO2_
                + 0.5
                * M_H2O
                * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                / (1.0 - xg_CO2_)
                - M_m_
                * (
                    -K5_ * P * y_CO2_
                    + 0.5
                    * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    / (1.0 - xg_CO2_)
                )
            )
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K5_
            * P
            * y_CO2_
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            - xg_CO2_
            * (-M_CO2 + M_H2O)
            / (M_CO2 * xg_CO2_ + M_H2O * (1.0 - xg_CO2_)) ** 2
            - 1 / (M_CO2 * xg_CO2_ + M_H2O * (1.0 - xg_CO2_))
        )
        / (
            -K5_
            * P
            * xg_CO2_
            * y_CO2_
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            + xg_CO2_ / (M_CO2 * xg_CO2_ + M_H2O * (1.0 - xg_CO2_))
        )
        ** 2
        + (
            -K5_
            * P
            * xg_CO2_
            * y_CO2_
            * (
                -K5_ * M_CO2 * P * y_CO2_
                + 0.5
                * M_H2O
                * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                / (1.0 - xg_CO2_)
                - M_m_
                * (
                    -K5_ * P * y_CO2_
                    + 0.5
                    * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    / (1.0 - xg_CO2_)
                )
            )
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K5_
            * P
            * y_CO2_
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K5_
            * P
            * xg_CO2_
            * y_CO2_
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            + xg_CO2_ / (M_CO2 * xg_CO2_ + M_H2O * (1.0 - xg_CO2_))
        )
        - (
            -((K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5)
            * (
                -K5_ * M_CO2 * P * y_CO2_
                + 0.5
                * M_H2O
                * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                / (1.0 - xg_CO2_)
                - M_m_
                * (
                    -K5_ * P * y_CO2_
                    + 0.5
                    * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    / (1.0 - xg_CO2_)
                )
            )
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
            / (
                (1.0 - xg_CO2_)
                * (
                    K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                    + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + M_m_
                    * (
                        -K5_ * P * xg_CO2_ * y_CO2_
                        - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -((K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5)
            / (
                K5_ * M_CO2 * P * xg_CO2_ * y_CO2_
                + M_H2O * (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                + M_m_
                * (
                    -K5_ * P * xg_CO2_ * y_CO2_
                    - (K4_ * P * y_H2O_ * (1.0 - xg_CO2_)) ** 0.5
                    + 1.0
                )
            )
            + (1.0 - xg_CO2_) / (M_CO2 * xg_CO2_ + M_H2O * (1.0 - xg_CO2_))
        )
    )
    return result


########################################
# SHOFe2 - inc. H2S and H2 - xg_S2 #####
########################################


def SHOFe2_HO_O2(xg_O2_, xg_S2_, constants):  # dmbHO_O
    """Evaluate differential of wtg_H - wtg_O with respect to xg_O2 for the S-H-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K4_,
        K6_,
        K7_,
        K8_,
        K9_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_OH,
        M_H2,
        M_H2O,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - 0.5
            * K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            * (
                -K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_H2O
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                - M_O2
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        + (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + 0.5
            * K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        - 0.111111111111111
        * (
            -3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                -K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_H2O
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                - M_O2
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                2.0 * K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                + 2.0
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.333333333333333
            * (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + 0.333333333333333
            * (
                2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
    )
    return result


def SHOFe2_HO_S2(xg_O2_, xg_S2_, constants):  # dmbHO_S
    """Evaluate differential of wtg_H - wtg_O with respect to xg_S2 for the S-H-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K4_,
        K6_,
        K7_,
        K8_,
        K9_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_OH,
        M_H2,
        M_H2O,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            -0.5
            * K13_
            * K7_
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K14_
            * K7_**2
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-1.0)
            * (xg_S2_ * y_S2_) ** 1.0
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            * (
                -0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K7_**2
                * M_H2S
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K7_
                * M_H2O
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_S2
                + 0.5
                * K7_
                * M_H2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                -0.5
                * K7_**2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - 0.5
                * K7_
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K4_
                * K7_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        + (
            0.5
            * K13_
            * K7_
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K14_
            * K7_**2
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-1.0)
            * (xg_S2_ * y_S2_) ** 1.0
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K4_
                * K7_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        - 0.111111111111111
        * (
            -3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                -0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K7_**2
                * M_H2S
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K7_
                * M_H2O
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_S2
                + 0.5
                * K7_
                * M_H2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K4_
                * K7_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            - 0.333333333333333
            * (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + 0.333333333333333
            * (
                2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K4_
                * K7_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_O2_
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
    )
    return result


def SHOFe2_HS_O2(xg_O2_, xg_S2_, constants):  # dmbHS_O
    """Evaluate differential of wtg_H - wtg_S with respect to xg_O2 for the S-H-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K4_,
        K6_,
        K7_,
        K8_,
        K9_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_OH,
        M_H2,
        M_H2O,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - 0.5
            * K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            * (
                -K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_H2O
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                - M_O2
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            * (
                -K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_H2O
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
                - M_O2
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_O2_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_O2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        / (
            -K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        + (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + 0.5
            * K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_O2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                0.5
                * K4_
                * P
                * y_H2O_
                * (-K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_ - 1)
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
    )
    return result


def SHOFe2_HS_S2(xg_O2_, xg_S2_, constants):  # dmbHS_S
    """Evaluate differential of wtg_H - wtg_S with respect to xg_S2 for the S-H-O-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K4_,
        K6_,
        K7_,
        K8_,
        K9_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_OH,
        M_H2,
        M_H2O,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            -0.5
            * K13_
            * K7_
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K14_
            * K7_**2
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-1.0)
            * (xg_S2_ * y_S2_) ** 1.0
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            * (
                -0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K7_**2
                * M_H2S
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K7_
                * M_H2O
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_S2
                + 0.5
                * K7_
                * M_H2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                -0.5
                * K7_**2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - 0.5
                * K7_
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            + (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K4_
                * K7_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            -K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            -0.5
            * K14_
            * K7_**2
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-1.0)
            * (xg_S2_ * y_S2_) ** 1.0
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            + K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            * (
                -0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K7_**2
                * M_H2S
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K7_
                * M_H2O
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_H2O
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                - M_S2
                + 0.5
                * K7_
                * M_H2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * xg_S2_
                    * y_H2S_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                - M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            ** 2
            - (
                0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 0.5
                * K7_**2
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-1.0)
                * (xg_S2_ * y_S2_) ** 1.0
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_**2
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        / (
            -K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        ** 2
        - (
            0.5
            * K14_
            * K7_**2
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-1.0)
            * (xg_S2_ * y_S2_) ** 1.0
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 2.0 * xg_S2_
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
        + (
            0.5
            * K13_
            * K7_
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K14_
            * K7_**2
            * P
            * y_H2O_**2
            * (xg_O2_ * y_O2_) ** (-1.0)
            * (xg_S2_ * y_S2_) ** 1.0
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * y_H2S_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - 1
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * xg_S2_
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
                + M_m_
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                * (
                    -0.25
                    * K4_
                    * K7_
                    * P
                    * y_H2O_**2
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        xg_S2_
                        * y_H2S_
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2
                    )
                    + 0.5
                    * K4_
                    * P
                    * y_H2O_
                    * (
                        -0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        - 1
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                )
            )
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            ** 2
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            * (
                -0.25
                * K4_
                * K7_
                * P
                * y_H2O_**2
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    xg_S2_
                    * y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2
                )
                + 0.5
                * K4_
                * P
                * y_H2O_
                * (
                    -0.5
                    * K6_
                    * xg_O2_
                    * y_O2_
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_SO2_)
                    - 1
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            * (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 1.0
                + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K4_
                * P
                * y_H2O_
                * (
                    M_H2O
                    * (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (
                            K4_
                            * P
                            * y_H2O_
                            * (
                                -K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                - xg_O2_
                                - xg_S2_
                                + 1.0
                            )
                            / (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                        )
                        ** 0.5
                    )
                )
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
            )
        )
        / (
            -K13_
            * P
            * y_H2O_
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                K1_
                * M_H2
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K14_
            * K7_
            * P
            * y_H2O_
            * (xg_O2_ * y_O2_) ** (-0.5)
            * (xg_S2_ * y_S2_) ** 0.5
            * (
                -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - xg_O2_
                - xg_S2_
                + 1.0
            )
            / (
                M_H2S
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - (
                K4_
                * P
                * y_H2O_
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            ** 0.5
            / (
                M_H2O
                * (
                    K4_
                    * P
                    * y_H2O_
                    * (
                        -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        - xg_O2_
                        - xg_S2_
                        + 1.0
                    )
                    / (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                ** 0.5
                + M_m_
                * (
                    1.0
                    - (
                        K4_
                        * P
                        * y_H2O_
                        * (
                            -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            - xg_O2_
                            - xg_S2_
                            + 1.0
                        )
                        / (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                    ** 0.5
                )
            )
            + (
                K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + K7_
                * M_H2S
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    y_H2S_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + M_H2O
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + M_H2
                * y_H2O_
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    - xg_O2_
                    - xg_S2_
                    + 1.0
                )
                / (
                    K1_
                    * y_H2_
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
        )
    )
    return result


##############
# CHOXFe #####
##############


def CHOXFe_CO_O2(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCO_O
    """Evaluate differential of wtg_C - wtg_O with respect to xg_O2 for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - 0.25
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 2.0
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -0.5 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOXFe_CO_CO(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCO_CO
    """Evaluate differential of wtg_C - wtg_O with respect to xg_CO for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - 0.25
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                2.0 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 1
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        / (
            -0.5 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - (
            -K11_ * P * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOXFe_CO_X(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCO_S
    """Evaluate differential of wtg_C - wtg_O with respect to xg_X for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.353553390593274
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K2_**2
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * y_CH4_**2
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                * (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X * xg_X_
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                0.5
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2O
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                - M_X
                + 0.5
                * M_H2
                * y_H2O_**1.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - 0.25
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.353553390593274
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                0.5
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2O
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                - M_X
                + 0.5
                * M_H2
                * y_H2O_**1.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        / (
            -0.5 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - 0.353553390593274
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
    )
    return result


def CHOXFe_CH_O2(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCH_O
    """Evaluate differential of wtg_C - wtg_H with respect to xg_O2 for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.75
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    - 0.125
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 1.0
                    / (K1_ * xg_O2_ * y_H2_)
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * xg_O2_ * y_CO_)
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.75
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.75
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    - 0.125
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 1.0
                    / (K1_ * xg_O2_ * y_H2_)
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * xg_O2_ * y_CO_)
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOXFe_CH_CO(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCH_CO
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + 1.0
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_**2 * y_CO_)
            + 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 1.0
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_**2 * y_CO_)
            - 0.707106781186548
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                1.0
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_
                * K4_
                * P
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOXFe_CH_X(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCH_S
    """Evaluate differential of wtg_C - wtg_H with respect to xg_X for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K2_**2
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * y_CH4_**2
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                * (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X * xg_X_
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                0.5
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2O
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                - M_X
                + 0.5
                * M_H2
                * y_H2O_**1.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            -1.0
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K13_
            * P
            * y_H2O_**1.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (K1_ * M_H2)
            - 0.353553390593274
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                0.5
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2O
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                - M_X
                + 0.5
                * M_H2
                * y_H2O_**1.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                -1.0
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                - 0.5
                * y_H2O_**1.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        - (
            1.0
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K13_
            * P
            * y_H2O_**1.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (K1_ * M_H2)
            + 0.353553390593274
            * K2_
            * xg_CO_
            * y_CO_
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_
                * y_CH4_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.707106781186548
            * (
                K3_
                * K4_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOXFe_CX_O2(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCS_O
    """Evaluate differential of wtg_C - wtg_X with respect to xg_O2 for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        xg_X_
        * (-K6_ * P * xg_X_ * y_X_ / M_X + wt_X / M_X)
        * (
            -0.25
            * K2_**2
            * M_CH4
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * y_CH4_**2
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K2_
            * M_CO2
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (xg_O2_ * y_CO2_)
            + 0.375
            * K2_
            * M_CH4
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * xg_O2_ * y_CH4_)
            - M_O2
            - 0.5
            * K3_
            * M_H2O
            * y_CH4_
            * y_H2O_ ** (-2.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    - 0.125
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 1.0
                    / (K1_ * xg_O2_ * y_H2_)
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * xg_CO_ * y_CO_)
            - 0.75
            * K3_
            * M_H2O
            * y_CH4_
            * y_H2O_ ** (-2.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            - 0.5
            * K3_
            * M_H2
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    - 0.125
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 1.0
                    / (K1_ * xg_O2_ * y_H2_)
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            - 0.5
            * K3_
            * M_H2
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
        )
        / (
            (
                -K6_ * P * xg_X_ * y_X_ / M_X
                + xg_X_
                / (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X * xg_X_
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                )
            )
            ** 2
            * (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - M_O2
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        - 0.125
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            - 0.125
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** (-0.5)
                            * (
                                -0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    0.5
                                    * K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / (xg_O2_ * y_CO2_)
                                    + 1
                                )
                                / (K3_ * y_CH4_)
                                + 0.75
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * xg_O2_ * y_CH4_)
                                - 0.125
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 1.0
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            + 0.5
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                        + 0.75
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 0.707106781186548
                    * M_H2O
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        - 0.707106781186548
                        * (
                            K3_
                            * K4_
                            * P
                            * y_CH4_
                            * y_H2O_ ** (-1.0)
                            * (xg_O2_ * y_O2_) ** 2.0
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            * (
                                2.0
                                * (
                                    -K2_
                                    * xg_CO_
                                    * y_CO_
                                    * y_H2O_**2.0
                                    * (xg_O2_ * y_O2_) ** (-2.0)
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                        / y_CO2_
                                        + xg_CO_
                                        + xg_O2_
                                        + xg_X_
                                        - 1.0
                                    )
                                    / (K3_ * y_CH4_)
                                    + 0.25
                                    * (
                                        1.0
                                        + y_H2O_
                                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                                        / (K1_ * y_H2_)
                                    )
                                    ** 2.0
                                )
                                ** 0.5
                                - 1.0
                                - y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            / (K2_ * xg_CO_ * y_CO_)
                        )
                        ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOXFe_CX_CO(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCS_CO
    """Evaluate differential of wtg_C - wtg_X with respect to xg_CO for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        xg_X_
        * (-K6_ * P * xg_X_ * y_X_ / M_X + wt_X / M_X)
        * (
            -0.25
            * K2_**2
            * M_CH4
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * y_CH4_**2
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
            - 0.25
            * K2_
            * M_CH4
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * y_CH4_)
            - M_CO
            - 1.0
            * K3_
            * M_H2O
            * y_CH4_
            * y_H2O_ ** (-2.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (K2_ * xg_CO_ * y_CO_)
            + 0.5
            * K3_
            * M_H2O
            * y_CH4_
            * y_H2O_ ** (-2.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * xg_CO_**2 * y_CO_)
            - 1.0
            * K3_
            * M_H2
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            + 0.5
            * K3_
            * M_H2
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
        )
        / (
            (
                -K6_ * P * xg_X_ * y_X_ / M_X
                + xg_X_
                / (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X * xg_X_
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                )
            )
            ** 2
            * (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - 0.707106781186548
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**1.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    * (
                        1.0
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
                            / (K3_ * y_CH4_)
                            - 0.5
                            * K2_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                        )
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        / (K2_ * xg_CO_ * y_CO_)
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_**2 * y_CO_)
                    )
                    / (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
    )
    return result


def CHOXFe_CX_X(xg_O2_, xg_CO_, xg_X_, constants):  # dmbCS_S
    """Evaluate differential of wtg_C - wtg_X with respect to xg_X for the C-H-O-X-Fe
    system with O2, CO, and X as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_X_ (float): Mole fraction of X in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_X,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K11_,
        K12_,
        K13_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_X_,
        M_C,
        M_H,
        M_O,
        M_X,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    result = (
        K6_
        * P
        * y_X_
        / (
            M_X
            * (
                -K6_ * P * xg_X_ * y_X_ / M_X
                + xg_X_
                / (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X * xg_X_
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                )
            )
        )
        - (-K6_ * P * xg_X_ * y_X_ / M_X + wt_X / M_X)
        * (
            K6_ * P * y_X_ / M_X
            - xg_X_
            * (
                0.5
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2O
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                - M_X
                + 0.5
                * M_H2
                * y_H2O_**1.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
            - 1
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        / (
            -K6_ * P * xg_X_ * y_X_ / M_X
            + xg_X_
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
        + (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + wt_C / M_C
        )
        * (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K2_**2
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**2.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_X_
                    - 1.0
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)) ** 2.0
            )
            ** (-0.5)
            / (
                K3_**2
                * y_CH4_**2
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                * (
                    K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 0.25
                    * K2_
                    * M_CH4
                    * xg_CO_
                    * y_CO_
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K3_
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 2.0
                    / (K3_ * y_CH4_)
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X * xg_X_
                    + 0.5
                    * K3_
                    * M_H2O
                    * y_CH4_
                    * y_H2O_ ** (-2.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * M_H2
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-1.0)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                0.353553390593274
                * K2_
                * M_H2O
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.353553390593274
                * K2_
                * M_m_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_
                    * y_CH4_
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            ** 2
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            * (
                0.5
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * M_H2O
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                - M_X
                + 0.5
                * M_H2
                * y_H2O_**1.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_X_
                        - 1.0
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (1.0 + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_))
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + 0.707106781186548
                * M_H2O
                * (
                    K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 0.707106781186548
                    * (
                        K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_X_
                                    - 1.0
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    ** 0.5
                    + 1.0
                )
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_X_
                                - 1.0
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X * xg_X_
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_X_
                            - 1.0
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
        )
        ** 2
    )
    return result


##########################################
# SCHOFe: O2, CO, S2 - linear H2O!!! #####
##########################################


def SCHOFe3_OCS_CO_O2(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCO_O
    """Evaluate differential of wtg_C - wtg_O with respect to xg_O2 for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            + 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 2.0 * K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
            + 0.75
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * xg_O2_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + 0.333333333333333
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.166666666666667
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - 1.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - 4.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
            - 0.75
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * xg_O2_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return value


def SCHOFe3_OCS_CO_CO(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCO_CO
    """Evaluate differential of wtg_C - wtg_O with respect to xg_CO for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - (
                2.0 * K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 1
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            + 1.0
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_**2
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + 0.333333333333333
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.166666666666667
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            -K11_ * P * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - 2.0
            * K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - 1.0
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_**2
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return value


def SCHOFe3_OCS_CO_S2(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCO_S
    """Evaluate differential of wtg_C - wtg_O with respect to xg_S2 for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        - 0.111111111111111
        * (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - 3.0
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / M_S
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        * (
            2.0
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                1.0
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
        )
        / (
            -0.333333333333333 * K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.666666666666667
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + 0.333333333333333
            * (
                2.0 * K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 2.0 * K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + xg_CO_
                + 2.0 * xg_O2_
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.166666666666667
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        ** 2
    )
    return value


def SCHOFe3_OCS_CH_O2(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCH_O
    """Evaluate differential of wtg_C - wtg_H with respect to xg_O2 for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5 * wt_H / M_H
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        * (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.75
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * xg_O2_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_O2_ * y_CO_)
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.75
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.75
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
            + 0.75
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * xg_O2_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.75
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * xg_O2_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_O2_ * y_CO_)
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
            - 0.75
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * xg_O2_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return value


def SCHOFe3_OCS_CH_CO(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCH_CO
    """Evaluate differential of wtg_C - wtg_H with respect to xg_CO for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5 * wt_H / M_H
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        * (
            0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + 1.0
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_**2 * y_CO_)
            + 1.0
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_**2 * y_CO_)
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1.0
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                + 1.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            + 1.0
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_**2
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 1.0
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_**2 * y_CO_)
            - 1.0
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_**2 * y_CO_)
            - 1.0
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_**2
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return value


def SCHOFe3_OCS_CH_S2(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCH_S
    """Evaluate differential of wtg_C - wtg_H with respect to xg_S2 for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5 * wt_H / M_H
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        * (
            1.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.25
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_S2_ * y_CO_)
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                1.0
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.25
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            + 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        ** 2
        - (
            -1.0
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.25
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_S2_ * y_CO_)
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2
            )
        )
        / (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - 0.5
            * K13_
            * K3_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-1.0)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K1_ * K2_ * M_H2 * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - 0.5
            * K3_
            * K4_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 2.0
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (
                K2_
                * xg_CO_
                * y_CO_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
    )
    return value


def SCHOFe3_OCS_CS_O2(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCS_O
    """Evaluate differential of wtg_C - wtg_S with respect to xg_O2 for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 0.375
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_O2_ * y_CO_)
            + 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            - 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_O2_ * y_H2S_)
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        / (xg_O2_ * y_CO2_)
                        + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + 1
                    )
                    / (K3_ * y_CH4_)
                    + 0.75
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * xg_O2_ * y_CH4_)
                    + 0.125
                    * (
                        -1.0
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        - 1.0
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                )
                + 0.5 * y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * xg_O2_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_O2_ * y_CO_)
            - 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_O2_)
            + 1.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_O2_)
            - (
                K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    + 3.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K2_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 0.375
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * xg_O2_ * y_CH4_)
                - K6_ * M_SO2 * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                - M_O2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * xg_O2_ * y_CO_ * y_H2_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 3.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.375
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4 * xg_O2_)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K2_
                * K5_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                - M_m_
                * (
                    -0.5
                    * K2_
                    * K5_
                    * P
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / xg_O2_
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        0.5
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / (xg_O2_ * y_H2S_)
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** (-0.5)
                        * (
                            -0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                0.5
                                * K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / (xg_O2_ * y_CO2_)
                                + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                                + 1
                            )
                            / (K3_ * y_CH4_)
                            + 0.75
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * xg_O2_ * y_CH4_)
                            + 0.125
                            * (
                                -1.0
                                * K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / (xg_O2_ * y_H2S_)
                                - 1.0
                                * y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * xg_O2_ * y_H2_)
                            )
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 1.0
                        )
                        + 0.5
                        * y_H2O_
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        / (K1_ * xg_O2_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                    - 0.75
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
                )
                - 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_O2_ * y_H2S_)
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K2_
                            * xg_CO_
                            * y_CO_
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            / (xg_O2_ * y_CO2_)
                            + K6_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                            + 1
                        )
                        / (K3_ * y_CH4_)
                        + 0.75
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * xg_O2_ * y_CH4_)
                        + 0.125
                        * (
                            -1.0
                            * K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / (xg_O2_ * y_H2S_)
                            - 1.0
                            * y_H2O_
                            * (P * xg_O2_ * y_O2_) ** (-0.5)
                            / (K1_ * xg_O2_ * y_H2_)
                        )
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                    )
                    + 0.5
                    * y_H2O_
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    / (K1_ * xg_O2_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.75
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_O2_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - 0.5
            * K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                        - 0.5
                        * K3_
                        * K4_
                        * P
                        * y_CH4_
                        * y_H2O_ ** (-1.0)
                        * (xg_O2_ * y_O2_) ** 2.0
                        * (P * xg_O2_ * y_O2_) ** (-0.5)
                        * (
                            -K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 2.0
                            * (
                                -K2_
                                * xg_CO_
                                * y_CO_
                                * y_H2O_**2.0
                                * (xg_O2_ * y_O2_) ** (-2.0)
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                * (
                                    K2_
                                    * xg_CO_
                                    * y_CO_
                                    * (P * xg_O2_ * y_O2_) ** 0.5
                                    / y_CO2_
                                    + K6_
                                    * xg_O2_
                                    * y_O2_
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    / y_SO2_
                                    + xg_CO_
                                    + xg_O2_
                                    + xg_S2_
                                    - 1.0
                                    + K6_
                                    * P
                                    * xg_O2_
                                    * y_O2_
                                    * (xg_CO_ * y_CO_) ** 3.0
                                    * (P * xg_S2_ * y_S2_) ** 0.5
                                    * (
                                        K2_
                                        * xg_CO_
                                        * y_CO_
                                        * (P * xg_O2_ * y_O2_) ** 0.5
                                    )
                                    ** (-2.0)
                                    / (K10_ * y_OCS_)
                                )
                                / (K3_ * y_CH4_)
                                + 0.25
                                * (
                                    K7_
                                    * y_H2O_
                                    * (xg_O2_ * y_O2_) ** (-0.5)
                                    * (xg_S2_ * y_S2_) ** 0.5
                                    / y_H2S_
                                    + 1.0
                                    + y_H2O_
                                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                                    / (K1_ * y_H2_)
                                )
                                ** 2.0
                            )
                            ** 0.5
                            - 1.0
                            - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        / (K2_ * xg_CO_ * y_CO_)
                    )
                    + 0.5
                    * K3_
                    * K4_
                    * M_H2O
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return value


def SCHOFe3_OCS_CS_CO(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCS_CO
    """Evaluate differential of wtg_C - wtg_S with respect to xg_CO for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        -(
            -1.0
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_**2 * y_CO_)
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            K11_ * P * y_CO_ / M_CO
            + 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            + K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.25
                * K2_**2
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + 1
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            1.0
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + 1
                    + 1.0
                    * K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * xg_CO_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                - 0.5
                * K2_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
            )
            * (
                -K2_
                * xg_CO_
                * y_CO_
                * y_H2O_**2.0
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                    + xg_CO_
                    + xg_O2_
                    + xg_S2_
                    - 1.0
                    + K6_
                    * P
                    * xg_O2_
                    * y_O2_
                    * (xg_CO_ * y_CO_) ** 3.0
                    * (P * xg_S2_ * y_S2_) ** 0.5
                    * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                    / (K10_ * y_OCS_)
                )
                / (K3_ * y_CH4_)
                + 0.25
                * (
                    K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 1.0
                    + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                ** 2.0
            )
            ** (-0.5)
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_**2 * y_CO_)
            - (
                1.0
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                + 1.0
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.25
                * K2_**2
                * M_CH4
                * xg_CO_**2
                * y_CO_**2
                * y_H2O_**1.0
                * (xg_O2_ * y_O2_) ** (-4.0)
                * (P * xg_O2_ * y_O2_) ** 1.0
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    4.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    - 2.0
                    * K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                / (
                    K3_**2
                    * y_CH4_**2
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - K2_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - 0.25
                * K2_
                * M_CH4
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                - M_CO
                - 1.0
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_ * y_H2S_)
                - 1.0
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
                - 1.0
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_**2 * y_CO_ * y_H2_)
                - 1.0
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_CO_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        + (
            -K11_ * P * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_**2
            * P
            * xg_CO_**2
            * y_CO_**2
            * y_H2O_**1.0
            * (xg_O2_ * y_O2_) ** (-4.0)
            * (P * xg_O2_ * y_O2_) ** 1.0
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                4.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                - 2.0
                * K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K3_**2
                * M_CH4
                * y_CH4_
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - 0.25
            * K12_
            * K2_
            * P
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K2_ * K5_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                - M_m_
                * (
                    -K2_ * K5_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    - 1.0
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + 1
                            + 1.0
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_CO_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        - 0.5
                        * K2_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                    / (K2_ * xg_CO_ * y_CO_)
                    + 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_**2 * y_CO_)
                )
                - 1.0
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + 1
                        + 1.0
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_CO_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    - 0.5
                    * K2_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_**2 * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - K2_
            * K5_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
    )
    return value


def SCHOFe3_OCS_CS_S2(xg_O2_, xg_CO_, xg_S2_, constants):  # dmbCS_S
    """Evaluate differential of wtg_C - wtg_S with respect to xg_S2 for the S-C-H-O-Fe
    system with linear H2O and O2, CO, and S2 as guessed species.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        xg_S2_ (float): Mole fraction of S2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_C,
        wt_H,
        wt_S,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        K5_,
        K6_,
        K7_,
        K8_,
        K9_,
        K10_,
        K11_,
        K12_,
        K13_,
        K14_,
        KD1_,
        KD2,
        y,
        y_H2_,
        y_O2_,
        y_H2O_,
        y_CO_,
        y_CO2_,
        y_CH4_,
        y_SO2_,
        y_S2_,
        y_H2S_,
        y_OCS_,
        M_C,
        M_H,
        M_O,
        M_S,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_OH,
        M_H2,
        M_H2O,
        M_CH4,
        M_S2,
        M_SO2,
        M_SO3,
        M_H2S,
        M_OCS,
        M_FeO,
        M_FeO15,
        M_m_,
        M_m_ox,
    ) = constants
    value = (
        (
            -0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        + (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + wt_C / M_C
        )
        * (
            0.5
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (
                K3_
                * M_CH4
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
            )
            + K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * K4_
                * M_m_
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2
            - (
                0.5
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            - (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
        )
        / (
            -K11_ * P * xg_CO_ * y_CO_ / M_CO
            - 0.25
            * K12_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (xg_O2_ * y_O2_) ** (-2.0)
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                K3_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            ** 2.0
            / (K3_ * M_CH4)
            - K2_
            * K5_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K2_ * K5_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (
                    -K2_ * K5_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + 1.0
                    - 0.5
                    * K3_
                    * K4_
                    * P
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                + 0.5
                * K3_
                * K4_
                * M_H2O
                * P
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
            )
            + (
                K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + xg_CO_
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - 0.25
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_S2_ * y_CO_)
            - 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            - 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        - (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + wt_S / M_S
        )
        * (
            0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -0.5
                * K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_H2S_)
                + 2.0
                * (
                    -0.5
                    * K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        0.5
                        * K6_
                        * xg_O2_
                        * y_O2_
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        / (xg_S2_ * y_SO2_)
                        + 1
                        + 0.5
                        * K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * xg_S2_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.125
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 1.0
                    / (xg_S2_ * y_H2S_)
                )
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** (-0.5)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            + 0.25
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * xg_S2_ * y_CO_)
            + 0.5 * K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / (M_S * xg_S2_)
            + 0.5
            * K9_
            * P**2.0
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            / (M_S * xg_S2_)
            - (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            * (
                -0.5
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (
                    K3_
                    * y_CH4_
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                )
                - 0.5
                * K6_
                * M_SO2
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                - M_S2
                - 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                - 0.25
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                - 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_)
                - 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                - 0.5
                * K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            ** 2
            - (
                0.5
                * K6_
                * xg_O2_
                * y_O2_
                * (P * xg_S2_ * y_S2_) ** 0.5
                / (xg_S2_ * y_SO2_)
                + 2.0
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -0.5
                    * K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / (xg_S2_ * y_H2S_)
                    + 2.0
                    * (
                        -0.5
                        * K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            0.5
                            * K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / (xg_S2_ * y_SO2_)
                            + 1
                            + 0.5
                            * K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * xg_S2_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.125
                        * K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 1.0
                        / (xg_S2_ * y_H2S_)
                    )
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** (-0.5)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.25
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * xg_S2_ * y_CO_ * y_H2S_)
                + 0.5
                * K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * xg_S2_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        / (
            -0.5
            * K14_
            * K3_
            * K7_
            * P
            * y_CH4_
            * y_H2O_ ** (-1.0)
            * (xg_O2_ * y_O2_) ** 1.5
            * (xg_S2_ * y_S2_) ** 0.5
            * (P * xg_O2_ * y_O2_) ** (-0.5)
            * (
                -K7_
                * y_H2O_
                * (xg_O2_ * y_O2_) ** (-0.5)
                * (xg_S2_ * y_S2_) ** 0.5
                / y_H2S_
                + 2.0
                * (
                    -K2_
                    * xg_CO_
                    * y_CO_
                    * y_H2O_**2.0
                    * (xg_O2_ * y_O2_) ** (-2.0)
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    * (
                        K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        + K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                        + xg_CO_
                        + xg_O2_
                        + xg_S2_
                        - 1.0
                        + K6_
                        * P
                        * xg_O2_
                        * y_O2_
                        * (xg_CO_ * y_CO_) ** 3.0
                        * (P * xg_S2_ * y_S2_) ** 0.5
                        * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                        / (K10_ * y_OCS_)
                    )
                    / (K3_ * y_CH4_)
                    + 0.25
                    * (
                        K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 1.0
                        + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    ** 2.0
                )
                ** 0.5
                - 1.0
                - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
            )
            / (K2_ * M_H2S * xg_CO_ * y_CO_)
            - K8_ * (xg_S2_ * y_S2_ / (xg_O2_ * y_O2_)) ** 0.5 / M_S
            - K9_ * P**2.0 * (xg_O2_ * y_O2_) ** 1.5 * (xg_S2_ * y_S2_) ** 0.5 / M_S
            + (
                K6_ * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + 2.0 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + K6_
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
            / (
                K2_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + 0.25
                * K2_
                * M_CH4
                * xg_CO_
                * y_CO_
                * (xg_O2_ * y_O2_) ** (-2.0)
                * (P * xg_O2_ * y_O2_) ** 0.5
                * (
                    K3_
                    * y_CH4_
                    * y_H2O_ ** (-1.0)
                    * (xg_O2_ * y_O2_) ** 2.0
                    * (P * xg_O2_ * y_O2_) ** (-0.5)
                    * (
                        -K7_
                        * y_H2O_
                        * (xg_O2_ * y_O2_) ** (-0.5)
                        * (xg_S2_ * y_S2_) ** 0.5
                        / y_H2S_
                        + 2.0
                        * (
                            -K2_
                            * xg_CO_
                            * y_CO_
                            * y_H2O_**2.0
                            * (xg_O2_ * y_O2_) ** (-2.0)
                            * (P * xg_O2_ * y_O2_) ** 0.5
                            * (
                                K2_
                                * xg_CO_
                                * y_CO_
                                * (P * xg_O2_ * y_O2_) ** 0.5
                                / y_CO2_
                                + K6_
                                * xg_O2_
                                * y_O2_
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                / y_SO2_
                                + xg_CO_
                                + xg_O2_
                                + xg_S2_
                                - 1.0
                                + K6_
                                * P
                                * xg_O2_
                                * y_O2_
                                * (xg_CO_ * y_CO_) ** 3.0
                                * (P * xg_S2_ * y_S2_) ** 0.5
                                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                                ** (-2.0)
                                / (K10_ * y_OCS_)
                            )
                            / (K3_ * y_CH4_)
                            + 0.25
                            * (
                                K7_
                                * y_H2O_
                                * (xg_O2_ * y_O2_) ** (-0.5)
                                * (xg_S2_ * y_S2_) ** 0.5
                                / y_H2S_
                                + 1.0
                                + y_H2O_
                                * (P * xg_O2_ * y_O2_) ** (-0.5)
                                / (K1_ * y_H2_)
                            )
                            ** 2.0
                        )
                        ** 0.5
                        - 1.0
                        - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                    )
                    / (K2_ * xg_CO_ * y_CO_)
                )
                ** 2.0
                / (K3_ * y_CH4_)
                + K6_ * M_SO2 * xg_O2_ * y_O2_ * (P * xg_S2_ * y_S2_) ** 0.5 / y_SO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_S2 * xg_S2_
                + 0.5
                * K3_
                * K7_
                * M_H2S
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 1.5
                * (xg_S2_ * y_S2_) ** 0.5
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_ * y_H2S_)
                + 0.5
                * K3_
                * M_H2O
                * y_CH4_
                * y_H2O_ ** (-2.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-0.5)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K2_ * xg_CO_ * y_CO_)
                + 0.5
                * K3_
                * M_H2
                * y_CH4_
                * y_H2O_ ** (-1.0)
                * (xg_O2_ * y_O2_) ** 2.0
                * (P * xg_O2_ * y_O2_) ** (-1.0)
                * (
                    -K7_
                    * y_H2O_
                    * (xg_O2_ * y_O2_) ** (-0.5)
                    * (xg_S2_ * y_S2_) ** 0.5
                    / y_H2S_
                    + 2.0
                    * (
                        -K2_
                        * xg_CO_
                        * y_CO_
                        * y_H2O_**2.0
                        * (xg_O2_ * y_O2_) ** (-2.0)
                        * (P * xg_O2_ * y_O2_) ** 0.5
                        * (
                            K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                            + K6_
                            * xg_O2_
                            * y_O2_
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            / y_SO2_
                            + xg_CO_
                            + xg_O2_
                            + xg_S2_
                            - 1.0
                            + K6_
                            * P
                            * xg_O2_
                            * y_O2_
                            * (xg_CO_ * y_CO_) ** 3.0
                            * (P * xg_S2_ * y_S2_) ** 0.5
                            * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5)
                            ** (-2.0)
                            / (K10_ * y_OCS_)
                        )
                        / (K3_ * y_CH4_)
                        + 0.25
                        * (
                            K7_
                            * y_H2O_
                            * (xg_O2_ * y_O2_) ** (-0.5)
                            * (xg_S2_ * y_S2_) ** 0.5
                            / y_H2S_
                            + 1.0
                            + y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                        )
                        ** 2.0
                    )
                    ** 0.5
                    - 1.0
                    - y_H2O_ * (P * xg_O2_ * y_O2_) ** (-0.5) / (K1_ * y_H2_)
                )
                / (K1_ * K2_ * xg_CO_ * y_CO_ * y_H2_)
                + K6_
                * M_OCS
                * P
                * xg_O2_
                * y_O2_
                * (xg_CO_ * y_CO_) ** 3.0
                * (P * xg_S2_ * y_S2_) ** 0.5
                * (K2_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5) ** (-2.0)
                / (K10_ * y_OCS_)
            )
        )
        ** 2
    )
    return value


#############
# COXFe #####
#############


def COXFe_XO_O2(xg_O2_, xg_CO_, constants):  # dmbXO_O
    """Evaluate differential of wtg_X - wtg_O with respect to xg_O2 for the C-O-X-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_X,
        wt_C,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        KD1_,
        KD2,
        y,
        y_X_,
        y_O2_,
        y_CO_,
        y_CO2_,
        M_C,
        M_X,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K4_
        * P
        * y_X_
        * (
            -0.5
            * K1_
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (xg_O2_ * y_CO2_)
            - 1
        )
        / (
            M_X
            * (
                -K4_
                * P
                * y_X_
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
                / M_X
                + (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
                / (
                    K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - xg_CO_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
        )
        + (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / M_X
            + wt_X / M_X
        )
        * (
            K4_
            * P
            * y_X_
            * (
                -0.5
                * K1_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 1
            )
            / M_X
            - (
                -0.5
                * K1_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 1
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -0.5
                * K1_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - M_O2
                - M_X
                * (
                    -0.5
                    * K1_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    - 1
                )
            )
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / M_X
            + (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - 0.25
        * (
            -2.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K1_
                * K2_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                + 0.5
                * K1_
                * K2_
                * M_m_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
            )
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            ** 2
            + 1.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                    )
                )
            )
            - (
                1.0
                * K1_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                + 2.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                2.0 * K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
            )
            * (
                -0.5
                * K1_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - M_O2
                - M_X
                * (
                    -0.5
                    * K1_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    - 1
                )
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - 0.5 * K3_ * P * xg_CO_ * y_CO_ / M_CO
            + 0.5
            * (
                2.0 * K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -2.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K1_
                * K2_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                + 0.5
                * K1_
                * K2_
                * M_m_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
            )
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            ** 2
            - 1.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                    )
                )
            )
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -2.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                2.0 * K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
    )
    return result


def COXFe_XO_CO(xg_O2_, xg_CO_, constants):  # dmbXO_C
    """Evaluate differential of wtg_X - wtg_O with respect to xg_CO for the C-O-X-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_X,
        wt_C,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        KD1_,
        KD2,
        y,
        y_X_,
        y_O2_,
        y_CO_,
        y_CO2_,
        M_C,
        M_X,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K4_
        * P
        * y_X_
        * (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
        / (
            M_X
            * (
                -K4_
                * P
                * y_X_
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
                / M_X
                + (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
                / (
                    K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - xg_CO_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
        )
        + (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / M_X
            + wt_X / M_X
        )
        * (
            K4_
            * P
            * y_X_
            * (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
            / M_X
            - (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -K1_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - M_CO
                - M_X * (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
            )
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / M_X
            + (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - 0.25
        * (
            -2.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            2.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K1_ * K2_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K1_ * K2_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            ** 2
            + 2.0
            * K1_
            * K2_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            + K3_ * P * y_CO_ / M_CO
            - (2.0 * K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -K1_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - M_CO
                - M_X * (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
            )
            * (
                2.0 * K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - 0.5 * K3_ * P * xg_CO_ * y_CO_ / M_CO
            + 0.5
            * (
                2.0 * K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -2.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K1_ * K2_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K1_ * K2_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            ** 2
            - 2.0
            * K1_
            * K2_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * y_CO_ / M_CO
        )
        / (
            -2.0
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + (
                2.0 * K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + xg_CO_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
    )
    return result


def COXFe_XC_O2(xg_O2_, xg_CO_, constants):  # dmbXC_O
    """Evaluate differential of wtg_X - wtg_C with respect to xg_O2 for the C-O-X-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_X,
        wt_C,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        KD1_,
        KD2,
        y,
        y_X_,
        y_O2_,
        y_CO_,
        y_CO2_,
        M_C,
        M_X,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K4_
        * P
        * y_X_
        * (
            -0.5
            * K1_
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (xg_O2_ * y_CO2_)
            - 1
        )
        / (
            M_X
            * (
                -K4_
                * P
                * y_X_
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
                / M_X
                + (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
                / (
                    K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - xg_CO_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
        )
        + (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / M_X
            + wt_X / M_X
        )
        * (
            K4_
            * P
            * y_X_
            * (
                -0.5
                * K1_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 1
            )
            / M_X
            - (
                -0.5
                * K1_
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - 1
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -0.5
                * K1_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - M_O2
                - M_X
                * (
                    -0.5
                    * K1_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    - 1
                )
            )
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / M_X
            + (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K1_
                * K2_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                + 0.5
                * K1_
                * K2_
                * M_m_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
            )
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            ** 2
            - 0.5
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                    )
                )
            )
        )
        / (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + (K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + xg_CO_)
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        - (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + wt_C / M_C
        )
        * (
            K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -0.5
                * K1_
                * K2_
                * M_CO2
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
                + 0.5
                * K1_
                * K2_
                * M_m_
                * P
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / xg_O2_
            )
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            ** 2
            + 0.5
            * K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * (
                    K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                    + M_m_
                    * (
                        -K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                        + 1.0
                    )
                )
            )
            - 0.5
            * K1_
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * y_CO2_
                * (
                    K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - xg_CO_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
            - (K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + xg_CO_)
            * (
                -0.5
                * K1_
                * M_CO2
                * xg_CO_
                * y_CO_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_CO2_)
                - M_O2
                - M_X
                * (
                    -0.5
                    * K1_
                    * xg_CO_
                    * y_CO_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_CO2_)
                    - 1
                )
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + (K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + xg_CO_)
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
    )
    return result


def COXFe_XC_CO(xg_O2_, xg_CO_, constants):  # dmbXC_C
    """Evaluate differential of wtg_X - wtg_C with respect to xg_CO for the C-O-X-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_CO_ (float): Mole fraction of CO in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_X,
        wt_C,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        KD1_,
        KD2,
        y,
        y_X_,
        y_O2_,
        y_CO_,
        y_CO2_,
        M_C,
        M_X,
        M_O,
        M_Fe,
        M_O2,
        M_CO,
        M_CO2,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K4_
        * P
        * y_X_
        * (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
        / (
            M_X
            * (
                -K4_
                * P
                * y_X_
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
                / M_X
                + (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
                / (
                    K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    + M_CO * xg_CO_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                        - xg_CO_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
        )
        + (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / M_X
            + wt_X / M_X
        )
        * (
            K4_
            * P
            * y_X_
            * (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
            / M_X
            - (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -K1_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - M_CO
                - M_X * (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
            )
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / M_X
            + (
                -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - xg_CO_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + wt_C / M_C
        )
        * (
            K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K1_ * K2_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K1_ * K2_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            ** 2
            + K1_
            * K2_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            + K3_ * P * y_CO_ / M_CO
            - (K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + 1)
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            - (K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + xg_CO_)
            * (
                -K1_ * M_CO2 * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                - M_CO
                - M_X * (-K1_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ - 1)
            )
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + (K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + xg_CO_)
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            * (
                -K1_ * K2_ * M_CO2 * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + K1_ * K2_ * M_m_ * P * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
            )
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            ** 2
            - K1_
            * K2_
            * P
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * y_CO_ / M_CO
        )
        / (
            -K1_
            * K2_
            * P
            * xg_CO_
            * y_CO_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                K1_ * K2_ * M_CO2 * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5
                + M_m_
                * (-K1_ * K2_ * P * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 + 1.0)
            )
            - K3_ * P * xg_CO_ * y_CO_ / M_CO
            + (K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_ + xg_CO_)
            / (
                K1_ * M_CO2 * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                + M_CO * xg_CO_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_CO_ * y_CO_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_CO2_
                    - xg_CO_
                    - xg_O2_
                    + 1.0
                )
            )
        )
    )
    return result


#############
# HOXFe #####
#############


def HOXFe_XO_O2(xg_O2_, xg_H2_, constants):  # dmbXO_O
    """Evaluate differential of wtg_X - wtg_O with respect to xg_O2 for the H-O-X-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_X,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        KD1_,
        KD2,
        y,
        y_X_,
        y_O2_,
        y_H2_,
        y_H2O_,
        M_H,
        M_X,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K4_
        * P
        * y_X_
        * (
            -0.5
            * K1_
            * xg_H2_
            * y_H2_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (xg_O2_ * y_H2O_)
            - 1
        )
        / (
            M_X
            * (
                -K4_
                * P
                * y_X_
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
                / M_X
                + (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
                / (
                    K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - xg_H2_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
        )
        + (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / M_X
            + wt_X / M_X
        )
        * (
            K4_
            * P
            * y_X_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 1
            )
            / M_X
            - (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 1
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - M_O2
                - M_X
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 1
                )
            )
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / M_X
            + (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -0.25
                * M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                + 0.25
                * M_m_
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
            )
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            ** 2
            - 0.25
            * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    M_H2O
                    * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                    )
                )
            )
            - wt_Fe
            * (
                -0.75
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                + 1.5
                * (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            * (
                0.5
                * KD1_ ** (2.0 * y)
                * KD2
                * y
                * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                * (1.0 - 2.0 * y)
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    xg_O2_
                    * (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    ** 2
                )
                - (
                    0.25 * KD1_ * (P * xg_O2_ * y_O2_) ** 0.25 / xg_O2_
                    + 1.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y**2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    / xg_O2_
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
                ** 2
            )
        )
        / (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        - (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.25
                * M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                + 0.25
                * M_m_
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
            )
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * xg_O2_
            )
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - M_O2
                - M_X
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 1
                )
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
            - (
                0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                + 2.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            + 0.25
            * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    M_H2O
                    * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                    )
                )
            )
        )
        / (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
    )
    return result


def HOXFe_XO_H2(xg_O2_, xg_H2_, constants):  # dmbXO_C
    """Evaluate differential of wtg_X - wtg_O with respect to xg_H2 for the H-O-X-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_X,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        KD1_,
        KD2,
        y,
        y_X_,
        y_O2_,
        y_H2_,
        y_H2O_,
        M_H,
        M_X,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K4_
        * P
        * y_X_
        * (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
        / (
            M_X
            * (
                -K4_
                * P
                * y_X_
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
                / M_X
                + (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
                / (
                    K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - xg_H2_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
        )
        + (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / M_X
            + wt_X / M_X
        )
        * (
            K4_
            * P
            * y_X_
            * (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
            / M_X
            - (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_H2
                - M_X * (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
            )
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / M_X
            + (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -0.5
                * M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            ** 2
            - 0.5
            * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    M_H2O
                    * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                    )
                )
            )
        )
        / (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        - (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + wt_O / M_O
            - wt_Fe
            * (
                1.5
                * (
                    KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                    + 2.0
                    * KD1_ ** (2.0 * y)
                    * KD2
                    * y
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                )
                / (
                    KD1_ ** (2.0 * y)
                    * KD2
                    * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    * (1.0 - 2.0 * y)
                    + 1.0
                )
                + 1.0
            )
            / (
                M_Fe
                * (
                    (
                        KD1_ * (P * xg_O2_ * y_O2_) ** 0.25
                        + 2.0
                        * KD1_ ** (2.0 * y)
                        * KD2
                        * y
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                    )
                    / (
                        KD1_ ** (2.0 * y)
                        * KD2
                        * (P * xg_O2_ * y_O2_) ** (0.5 * y)
                        * (1.0 - 2.0 * y)
                        + 1.0
                    )
                    + 1.0
                )
            )
        )
        * (
            -K1_
            * y_H2_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                y_H2O_
                * (
                    K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - xg_H2_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
            + (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            ** 2
            - (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * xg_O2_
            )
            * (
                -K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_H2
                - M_X * (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    M_H2O
                    * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                    )
                )
            )
        )
        / (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + (
                K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + 2.0 * xg_O2_
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
    )
    return result


def HOXFe_XH_O2(xg_O2_, xg_H2_, constants):  # dmbXC_O
    """Evaluate differential of wtg_X - wtg_H with respect to xg_O2 for the H-O-X-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_X,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        KD1_,
        KD2,
        y,
        y_X_,
        y_O2_,
        y_H2_,
        y_H2O_,
        M_H,
        M_X,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K4_
        * P
        * y_X_
        * (
            -0.5
            * K1_
            * xg_H2_
            * y_H2_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (xg_O2_ * y_H2O_)
            - 1
        )
        / (
            M_X
            * (
                -K4_
                * P
                * y_X_
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
                / M_X
                + (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
                / (
                    K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - xg_H2_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
        )
        - (
            -((K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5)
            * (
                -0.25
                * M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                + 0.25
                * M_m_
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
            )
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            ** 2
            - 0.25
            * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    M_H2O
                    * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                    )
                )
            )
        )
        / (
            -K3_ * P * xg_H2_ * y_H2_ / M_H2
            - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ + xg_H2_)
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        + (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / M_X
            + wt_X / M_X
        )
        * (
            K4_
            * P
            * y_X_
            * (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 1
            )
            / M_X
            - (
                -0.5
                * K1_
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - 1
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - M_O2
                - M_X
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 1
                )
            )
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / M_X
            + (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -K3_ * P * xg_H2_ * y_H2_ / M_H2
            - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            -0.5
            * K1_
            * xg_H2_
            * y_H2_
            * (P * xg_O2_ * y_O2_) ** 0.5
            / (
                xg_O2_
                * y_H2O_
                * (
                    K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - xg_H2_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
            + (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.25
                * M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
                + 0.25
                * M_m_
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_O2_
            )
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            ** 2
            - (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ + xg_H2_)
            * (
                -0.5
                * K1_
                * M_H2O
                * xg_H2_
                * y_H2_
                * (P * xg_O2_ * y_O2_) ** 0.5
                / (xg_O2_ * y_H2O_)
                - M_O2
                - M_X
                * (
                    -0.5
                    * K1_
                    * xg_H2_
                    * y_H2_
                    * (P * xg_O2_ * y_O2_) ** 0.5
                    / (xg_O2_ * y_H2O_)
                    - 1
                )
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
            + 0.25
            * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_O2_
                * (
                    M_H2O
                    * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                    )
                )
            )
        )
        / (
            -K3_ * P * xg_H2_ * y_H2_ / M_H2
            - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ + xg_H2_)
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
    )
    return result


def HOXFe_XH_H2(xg_O2_, xg_H2_, constants):  # dmbXC_C
    """Evaluate differential of wtg_X - wtg_H with respect to xg_H2 for the H-O-X-Fe
    system.

    Args:
        xg_O2_ (float): Mole fraction of O2 in the gas
        xg_H2_ (float): Mole fraction of H2 in the gas
        constants (list): Constants required to evaluate differential

    Returns:
        float: Value of differential
    """
    (
        P,
        wt_O,
        wt_X,
        wt_H,
        wt_Fe,
        K1_,
        K2_,
        K3_,
        K4_,
        KD1_,
        KD2,
        y,
        y_X_,
        y_O2_,
        y_H2_,
        y_H2O_,
        M_H,
        M_X,
        M_O,
        M_Fe,
        M_O2,
        M_H2,
        M_H2O,
        M_FeO,
        M_FeO15,
        M_m_,
    ) = constants
    result = (
        -K4_
        * P
        * y_X_
        * (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
        / (
            M_X
            * (
                -K4_
                * P
                * y_X_
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
                / M_X
                + (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
                / (
                    K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    + M_H2 * xg_H2_
                    + M_O2 * xg_O2_
                    + M_X
                    * (
                        -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                        - xg_H2_
                        - xg_O2_
                        + 1.0
                    )
                )
            )
        )
        + (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / M_X
            + wt_X / M_X
        )
        * (
            K4_
            * P
            * y_X_
            * (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
            / M_X
            - (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            - (
                -K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_H2
                - M_X * (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
            )
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
        )
        / (
            -K4_
            * P
            * y_X_
            * (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / M_X
            + (
                -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - xg_H2_
                - xg_O2_
                + 1.0
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
        - (
            -K3_ * P * y_H2_ / M_H2
            - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            ** 2
            - 0.5
            * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    M_H2O
                    * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                    )
                )
            )
        )
        / (
            -K3_ * P * xg_H2_ * y_H2_ / M_H2
            - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ + xg_H2_)
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        - (
            -K3_ * P * xg_H2_ * y_H2_ / M_H2
            - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + 0.5 * wt_H / M_H
        )
        * (
            K3_ * P * y_H2_ / M_H2
            + (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            * (
                -0.5
                * M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
                + 0.5
                * M_m_
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                / xg_H2_
            )
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            ** 2
            - (K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ + 1)
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            - (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ + xg_H2_)
            * (
                -K1_ * M_H2O * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                - M_H2
                - M_X * (-K1_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ - 1)
            )
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
            ** 2
            + 0.5
            * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                xg_H2_
                * (
                    M_H2O
                    * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                    + M_m_
                    * (
                        1.0
                        - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                        ** 0.5
                    )
                )
            )
        )
        / (
            -K3_ * P * xg_H2_ * y_H2_ / M_H2
            - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
            / (
                M_H2O
                * (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5) ** 0.5
                + M_m_
                * (
                    1.0
                    - (K1_ * K2_ * P * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5)
                    ** 0.5
                )
            )
            + (K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_ + xg_H2_)
            / (
                K1_ * M_H2O * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                + M_H2 * xg_H2_
                + M_O2 * xg_O2_
                + M_X
                * (
                    -K1_ * xg_H2_ * y_H2_ * (P * xg_O2_ * y_O2_) ** 0.5 / y_H2O_
                    - xg_H2_
                    - xg_O2_
                    + 1.0
                )
            )
        )
        ** 2
    )
    return result
